
from celescope.__init__ import HELP_DICT, HELP_INFO_DICT
from celescope.tools.step import Step, s_common
from celescope.tools import utils
import pandas as pd
import numpy as np


def get_opts_count_tag(parser, sub_program):
    if sub_program:
        parser.add_argument("--read_count_file", help="Tag read count file.", required=True)
        parser.add_argument("--match_dir", help=HELP_DICT['match_dir'])
        parser.add_argument("--matrix_dir", help=HELP_DICT['matrix_dir'])
        parser.add_argument("--tsne_file", help=HELP_DICT['tsne_file'])

        s_common(parser)

class Count_tag(Step):
    """
    ## Features
    - Assign tag to each cell barcode and summarize.

    ## Output

    - `{sample}_umi_tag.tsv` 

        `first column` cell barcode  
        `last column`  assigned tag  
        `columns between first and last` UMI count for each tag 

    - `{sample}_tsne_tag.tsv` it is `{sample}_umi_tag.tsv` with t-SNE coordinates, gene_counts and cluster infomation
    """
    
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)
        self.read_count_file = args.read_count_file

        # vals
        self.mapped_read = 0
        self.mapped_read_in_cell = 0
        self.df_UMI_cell = None
        self.df_tsne_tag = None

        # read
        self.df_read_count = pd.read_csv(self.read_count_file, sep="\t", index_col=0)

        if utils.check_arg_not_none(args, 'match_dir'):
            match_dict = utils.parse_match_dir(args.match_dir)
            self.match_barcode = match_dict['match_barcode']
            self.n_match_barcode = match_dict['n_match_barcode']
            self.tsne_file = match_dict['tsne_coord']
            self.matrix_dir = match_dict['matrix_dir']
        elif utils.check_arg_not_none(args, 'matrix_dir'):
            self.match_barcode, self.n_match_barcode = utils.get_barcode_from_matrix_dir(args.matrix_dir)
            self.tsne_file = args.tsne_file
            self.matrix_dir = args.matrix_dir
        else:
            raise ValueError("--match_dir or --matrix_dir is required.")

        # out files
        self.UMI_tag_file = f'{self.outdir}/{self.sample}_umi_tag.tsv'
        self.tsne_tag_file = f'{self.outdir}/{self.sample}_tsne_tag.tsv'

    def add_seq_metrics(self):

        self.add_metric(
            name=HELP_INFO_DICT['matched_barcode_number']['display'],
            value=self.n_match_barcode,
            help_info=HELP_INFO_DICT['matched_barcode_number']['info'],
        )

        self.add_metric(
            name='Mapped Reads in Cells',
            value=self.mapped_read_in_cell,
            total=self.mapped_read,
            help_info="Mapped reads with scRNA-Seq cell barcode"
        )

        UMIs = self.df_UMI_cell.apply(sum, axis=1)
        umi_median = round(np.median(UMIs), 2)
        umi_mean = round(np.mean(UMIs), 2)
        self.add_metric(
            name='Median UMI per Cell',
            value=umi_median,
            help_info="Median UMI per scRNA-Seq cell barcode"
        )
        self.add_metric(
            name='Mean UMI per Cell',
            value=umi_mean,
            help_info="Mean UMI per scRNA-Seq cell barcode"
        )

    def get_df_UMI_cell(self):
        self.mapped_read = int(self.df_read_count['read_count'].sum())
        df_read_count_in_cell = self.df_read_count[self.df_read_count.index.isin(self.match_barcode)]
        self.mapped_read_in_cell = int(df_read_count_in_cell['read_count'].sum())
        tag_name = df_read_count_in_cell.columns[0]
        df_UMI_in_cell = df_read_count_in_cell.reset_index().groupby([
            'barcode', tag_name]).agg({'UMI': 'count'})
        df_UMI_in_cell = df_UMI_in_cell.reset_index()
        df_UMI_in_cell = df_UMI_in_cell.pivot(
            index='barcode', columns=tag_name, values='UMI')
        df_cell = pd.DataFrame(index=self.match_barcode)
        df_UMI_cell = pd.merge(
            df_cell,
            df_UMI_in_cell,
            how="left",
            left_index=True,
            right_index=True
        )

        # fillna
        df_UMI_cell.fillna(0, inplace=True)
        self.df_UMI_cell = df_UMI_cell.astype(int)

    
    def get_df_tsne_tag(self):
        df_tsne = pd.read_csv(self.tsne_file, sep="\t", index_col=0)
        
        self.df_tsne_tag = pd.merge(
            df_tsne,
            self.df_UMI_cell,
            how="left",
            left_index=True,
            right_index=True)


    def write_files(self):
        self.df_UMI_cell.to_csv(self.UMI_tag_file, sep="\t")
        self.df_tsne_tag.to_csv(self.tsne_tag_file, sep="\t")

    def run(self):
        self.get_df_UMI_cell()
        self.get_df_tsne_tag()
        self.add_seq_metrics()
        self.write_files()


def count_tag(args):
    
    with Count_tag(args, display_title="Cells") as runner:
        runner.run()



import subprocess
import datetime

from celescope.__init__ import __version__
from celescope.tools.utils import add_log

ENV_NAME = f'celescope{__version__}'
CONDA_ROOT = '/SGRNJ/Public/Software/conda_env/'
CHANGELOG = 'docs_template/CHANGELOG.md'
TIMEFORMAT = '%Y-%m-%d'
TIME = datetime.datetime.now().strftime(TIMEFORMAT)

@add_log
def create_conda():
    cmd = f"""
    set -eo pipefail
    conda create -n {ENV_NAME}
    source activate {ENV_NAME}
    conda install -y --file conda_pkgs.txt

    pip install -i https://pypi.mirrors.ustc.edu.cn/simple/ -r requirements.txt
    python setup.py install
    """
    print(cmd)
    subprocess.check_call(cmd, shell=True)


@add_log
def lint_code():
    cmd = """
    set -eo pipefail
    celescope -h
    pip install -i https://pypi.mirrors.ustc.edu.cn/simple/ pylint
    # lint
        # W1618 (no-absolute-import)
        # E1101 (no-member)
        # W1633 (round-builtin)
        # W1619 (old-division)
        # W0105 (String statement has no effect)
        # W0511 TODO!
        # E1130 bad operand type for unary ~: _isnan (invalid-unary-operand-type)
    pylint --disable=all --enable=E,W --disable=W1618,E1101,W1633,W1619,W0105,W0511,E1130 --jobs=16 celescope
    """
    print(cmd)
    subprocess.check_call(cmd, shell=True)


@add_log
def zip_wdl():
    cmd = "cd wdl/ && zip -r wdl.zip ./*"
    print(cmd)
    subprocess.check_call(cmd, shell=True)


@add_log
def test_wdl():
    cmd = (
        "cd /SGRNJ03/randd/user/zhouyiqi/temp/wdl; "
        "sh /SGRNJ/Database/script/pipe/develop/dev_CeleScope/wdl/rna/local/run.sh "
    )
    print(cmd)
    subprocess.check_call(cmd, shell=True)


@add_log
def multi_test():
    cmd = (
        "cd /SGRNJ03/randd/user/zhouyiqi/github/celescope_test_script; "
        "pytest -s test_multi.py "
    )
    print(cmd)
    subprocess.check_call(cmd, shell=True)

@add_log
def generate_docs():
    cmd = (
        "python scripts/generate_docs.py"
    )
    print(cmd)
    subprocess.check_call(cmd, shell=True)

@add_log
def modify_changelog():
    header = f"## [unreleased] - {TIME}\n "
    lines = [header]
    with open(CHANGELOG, 'r') as f:
        for line in f:
            if line.find("unreleased") != -1:
                line = f'## [{__version__}] - {TIME}\n'
            lines.append(line)
    
    with open(CHANGELOG, 'w') as f:
        for line in lines:
            f.write(line)



if __name__ == '__main__':
    modify_changelog()
    generate_docs()
    lint_code()
    zip_wdl()
    create_conda()
    multi_test()
    test_wdl()



STEPS = [
    'mkref',
    'sample', 'barcode', 'cutadapt', 'consensus', 'star', 'featureCounts',
    'target_metrics', 'variant_calling', 'filter_snp', 'analysis_snp'
]
__ASSAY__ = 'snp'
IMPORT_DICT = {
    'star': 'celescope.rna'
}

PANEL = {'lung_1', 'blood_1'}



import os
import glob
import re
import sys
import unittest
from collections import Counter, defaultdict
from itertools import combinations, product

import pysam
from xopen import xopen

from celescope.tools import utils
from celescope.tools.__init__ import PATTERN_DICT
from celescope.__init__ import ROOT_PATH, HELP_DICT
from celescope.tools.step import Step, s_common

MIN_T = 10


class Chemistry():
    """
    Auto detect chemistry from R1-read
    """

    def __init__(self, fq1, assay=None):
        '''
        'scopeV2.0.1': 'C8L16C8L16C8L1U8T18'
        'scopeV2.1.1': 'C8L16C8L16C8L1U12T18'
        'scopeV2.2.1': 'C8L16C8L16C8L1U12T18' with 4 types of linkers
        'scopeV3.0.1': 'C9L16C9L16C9L1U12T18' with 4 types of linkers
        '''
        self.fq1 = fq1
        self.assay = assay
        self.fq1_list = fq1.split(',')
        self.n_read = 10000

        self.pattern_dict_v2, * \
            _, self.linker_1_v2_set_list, self.linker_1_v2_mismatch_list = Barcode.parse_chemistry('scopeV2.1.1')
        self.pattern_dict_v2, * \
            _, self.linker_4_v2_set_list, self.linker_4_v2_mismatch_list = Barcode.parse_chemistry('scopeV2.2.1')
        self.pattern_dict_v3, *_, self.linker_v3_set_list, self.linker_v3_mismatch_list = Barcode.parse_chemistry('scopeV3.0.1')
        self.pattern_dict_flv, *_, self.linker_flv_set_list, self.linker_flv_mismatch_list = Barcode.parse_chemistry('flv')
        self.pattern_dict_flv_rna, *_, self.linker_flv_rna_set_list, self.linker_flv_rna_mismatch_list = Barcode.parse_chemistry('flv_rna')


    @utils.add_log
    def check_chemistry(self):
        """check chemistry in the fq1_list"""
        if self.assay == 'bulk_vdj':
            return ['bulk_vdj'] * len(self.fq1_list)
        if self.assay == 'bulk_rna':
            return ['bulk_rna'] * len(self.fq1_list)
        chemistry_list = []
        for fastq1 in self.fq1_list:
            self.check_chemistry.logger.info(fastq1)
            chemistry = self.get_chemistry(fastq1)
            chemistry_list.append(chemistry)
        if len(set(chemistry_list)) != 1:
            Chemistry.check_chemistry.logger.warning('multiple chemistry found!' + str(chemistry_list))
        return chemistry_list


    def seq_chemistry(self, seq):
        """
        Returns: chemistry or None

        >>> runner = Chemistry("fake_fq1_string")
        >>> seq = "TCGACTGTCATCCACGTGCTTGAGATTCTAGGATTCAGCATGCGGCTACGTGCACGAGACATATCAATGGGTTTTCTTGTTGCTTTTTTTTTTTTTTTTTTTTTTTT"
        >>> runner.seq_chemistry(seq)
        'scopeV3.0.1'

        >>> seq = "GTCGTAGAATCCACGTGCTTGAGACTCAATGATCAGCATGCGGCTACGGCGATTAACGTTGAATGTTTTTTTTTTTTTTTTTTTTT"
        >>> runner.seq_chemistry(seq)
        'scopeV2.0.1'

        >>> seq = "NCAGATTC" + "ATCCACGTGCTTGAGA" + "GTACGCAA" + "TCAGCATGCGGCTACG" + "CTGAGCCA" + "C" + "TCCGAAGCCCAT" + "TTTTTTTTTTTTTTTTTTTTTTTTTTATTGC"
        >>> runner.seq_chemistry(seq)
        'scopeV2.1.1'

        >>> seq = "NCAGATTC" + "TCGGTGACAGCCATAT" + "GTACGCAA" + "CGTAGTCAGAAGCTGA" + "CTGAGCCA" + "C" + "TCCGAAGCCCAT" + "TTTTTTTTTTTTTTTTTTTTTTTTTTATTGC"
        >>> runner.seq_chemistry(seq)
        'scopeV2.2.1'

        """

        # check flv_rna first. otherwise it may be considered as scopeV2.1.1 and scopeV2.2.1
        linker_flv_rna = Barcode.get_seq_str(seq, self.pattern_dict_flv_rna["L"])
        bool_valid, _, _ = Barcode.check_seq_mismatch(
            [linker_flv_rna], self.linker_flv_rna_set_list, self.linker_flv_rna_mismatch_list)
        if bool_valid:
            return "flv_rna"


        linker_v2 = Barcode.get_seq_str(seq, self.pattern_dict_v2["L"])
        bool_valid, _, _ = Barcode.check_seq_mismatch(
            [linker_v2], self.linker_1_v2_set_list, self.linker_1_v2_mismatch_list)
        if bool_valid:
            if seq[65:69] == "TTTT":
                return "scopeV2.0.1"
            else:
                return "scopeV2.1.1"

        linker_v3 = Barcode.get_seq_str(seq, self.pattern_dict_v3["L"])
        bool_valid, _, _ = Barcode.check_seq_mismatch(
            [linker_v3], self.linker_v3_set_list, self.linker_v3_mismatch_list)
        if bool_valid:
            return "scopeV3.0.1"

        linker_v2 = Barcode.get_seq_str(seq, self.pattern_dict_v2["L"])
        bool_valid, _, _ = Barcode.check_seq_mismatch(
            [linker_v2], self.linker_4_v2_set_list, self.linker_4_v2_mismatch_list)
        if bool_valid:
            return "scopeV2.2.1"

        linker_flv = Barcode.get_seq_str(seq, self.pattern_dict_flv["L"])
        bool_valid, _, _ = Barcode.check_seq_mismatch(
            [linker_flv], self.linker_flv_set_list, self.linker_flv_mismatch_list)
        if bool_valid:
            return "flv"

        return

    @utils.add_log
    def get_chemistry(self, fq1):
        results = defaultdict(int)

        with pysam.FastxFile(fq1) as fh:
            for _ in range(self.n_read):
                entry = fh.__next__()
                seq = entry.sequence
                chemistry = self.seq_chemistry(seq)
                if chemistry:
                    results[chemistry] += 1
        # if it is 0, then no other linker types
        if results["scopeV2.2.1"] != 0:
            results["scopeV2.2.1"] += results["scopeV2.1.1"]
        sorted_counts = sorted(results.items(), key=lambda x: x[1], reverse=True)
        self.get_chemistry.logger.info(sorted_counts)

        chemistry, read_counts = sorted_counts[0][0], sorted_counts[0][1]
        percent = float(read_counts) / self.n_read
        if percent < 0.5:
            self.get_chemistry.logger.warning("Valid chemistry read counts percent < 0.5")
        if percent < 0.1:
            self.get_chemistry.logger.error("Valid chemistry read counts percent < 0.1")
            raise Exception(
                'Auto chemistry detection failed! ' + HELP_DICT['chemistry']
            )
        Chemistry.get_chemistry.logger.info(f'chemistry: {chemistry}')

        return chemistry

    @staticmethod
    def get_whitelist(chemistry):
        # returns: [bclists]
        repeat = 3
        root_dir = f'{ROOT_PATH}/data/chemistry/{chemistry}'
        bclist = f'{root_dir}/bclist'
        bclist1 = f'{root_dir}/bclist1'
        if os.path.exists(bclist1):
            return [f'{root_dir}/bclist{i}' for i in range(1,repeat+1)]
        elif os.path.exists(bclist):
            return [bclist] * repeat
        else:
            sys.exit(f'No bclist found for chemistry {chemistry} under {root_dir}')


class Barcode(Step):
    """
    ## Features

    - Demultiplex barcodes.
    - Filter invalid R1 reads, which includes:
        - Reads without linker: the mismatch between linkers and all linkers in the whitelist is greater than 2.  
        - Reads without correct barcode: the mismatch between barcodes and all barcodes in the whitelist is greater than 1.  
        - Reads without polyT: the number of T bases in the defined polyT region is less than 10.
        - Low quality reads: low sequencing quality in barcode and UMI regions.

    ## Output

    - `01.barcode/{sample}_2.fq(.gz)` Demultiplexed R2 reads. Barcode and UMI are contained in the read name. The format of 
    the read name is `{barcode}_{UMI}_{read ID}`.
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.fq1_list = args.fq1.split(",")
        self.fq2_list = args.fq2.split(",")
        self.fq_number = len(self.fq1_list)
        if self.fq_number != len(self.fq2_list):
            raise Exception('fastq1 and fastq2 do not have same file number!')
        if args.chemistry == 'auto':
            ch = Chemistry(args.fq1, self.assay)
            self.chemistry_list = ch.check_chemistry()
        else:
            self.chemistry_list = [args.chemistry] * self.fq_number
        self.barcode_corrected_num = 0
        self.linker_corrected_num = 0
        self.total_num = 0
        self.clean_num = 0
        self.no_polyT_num = 0
        self.lowQual_num = 0
        self.no_linker_num = 0
        self.no_barcode_num = 0
        self.barcode_qual_Counter = Counter()
        self.umi_qual_Counter = Counter()
        self.pattern = args.pattern
        self.linker = args.linker
        self.whitelist = args.whitelist
        self.lowNum = args.lowNum
        self.lowQual = args.lowQual
        self.filterNoPolyT = args.filterNoPolyT
        self.allowNoLinker = args.allowNoLinker
        self.nopolyT = args.nopolyT  # true == output nopolyT reads
        self.noLinker = args.noLinker
        self.output_R1 = args.output_R1
        self.bool_flv = False
        self.wells = args.wells


        # flv_trust4, flv_CR
        if self.assay in ('flv_CR', 'flv_trust4'):
            self.bool_flv = True
            self.barcode_read_Counter = Counter()
            if self.assay == 'flv_trust4':
                if args.match_dir == 'None':
                    raise FileNotFoundError('Match_dir required when running flv_trust4')
                self.match_barcodes = set(utils.get_barcode_from_match_dir(args.match_dir)[0]) # barcode set of flv_rna.
                self.match_num = 0 # record read number match with flv_rna.
                self.match_cbs = set() # record barcode number match with flv_rna.

        # out file
        if args.gzip:
            self.suffix = ".gz"
        else:
            self.suffix = ""
        self.out_fq2 = f'{self.out_prefix}_2.fq{self.suffix}'
        self.out_fq1 = f'{self.out_prefix}_1.fq{self.suffix}'
        if self.nopolyT:
            self.nopolyT_1 = f'{self.out_prefix}_noPolyT_1.fq'
            self.nopolyT_2 = f'{self.out_prefix}_noPolyT_2.fq'
        if self.noLinker:
            self.noLinker_1 = f'{self.out_prefix}_noLinker_1.fq'
            self.noLinker_2 = f'{self.out_prefix}_noLinker_2.fq'

        self.open_files()

    @staticmethod
    def get_seq_str_no_exception(seq, sub_pattern_dict):
        """get subseq with intervals in arr and concatenate"""
        return ''.join([seq[item[0]: item[1]] for item in sub_pattern_dict])

    @staticmethod
    def get_seq_str(seq, sub_pattern_dict):
        """
        Get subseq with intervals in arr and concatenate

        Args:
            seq: str
            sub_pattern_dict: [[0, 8], [24, 32], [48, 56]]

        Returns:
            str
            if sequence length is not enough, return ""

        >>> sub_pattern_dict = [[0, 8]]
        >>> seq = "A" * 7
        >>> Barcode.get_seq_str(seq, sub_pattern_dict)
        ""
        >>> seq = "A" * 8
        >>> Barcode.get_seq_str(seq, sub_pattern_dict)
        'AAAAAAAA'
        """
        seq_len = len(seq)
        ans = []
        for item in sub_pattern_dict:
            start, end = item[0], item[1]
            if end > seq_len:
                raise IndexError(f"sequence length is not enough in R1 read: {seq}")
            else:
                ans.append(seq[start:end])
        return ''.join(ans)

    @staticmethod
    def get_seq_list(seq, pattern_dict, abbr):
        """
        >>> pattern_dict = Barcode.parse_pattern("C2L3C2")
        >>> seq = "AAGGGTT"
        >>> Barcode.get_seq_list(seq, pattern_dict, "C")
        ['AA', 'TT']
        """
        
        return [seq[item[0]: item[1]] for item in pattern_dict[abbr]]

    @staticmethod
    def parse_pattern(pattern):
        """
        >>> pattern_dict = Barcode.parse_pattern("C8L16C8L16C8L1U12T18")
        >>> pattern_dict['C']
        [[0, 8], [24, 32], [48, 56]]
        >>> pattern_dict['L']
        [[8, 24], [32, 48], [56, 57]]
        """
        pattern_dict = defaultdict(list)
        p = re.compile(r'([CLUNT])(\d+)')
        tmp = p.findall(pattern)
        if not tmp:
            sys.exit()(f'Invalid pattern: {pattern}')
        start = 0
        for item in tmp:
            end = start + int(item[1])
            pattern_dict[item[0]].append([start, end])
            start = end
        return pattern_dict

    @staticmethod
    def get_abbr_len(pattern_dict, abbr):
        """
        >>> pattern_dict = Barcode.parse_pattern("C8L16C8L16C8L1U12T18")
        >>> Barcode.get_abbr_len(pattern_dict, 'C')
        24
        >>> Barcode.get_abbr_len(pattern_dict, 'L')
        33
        """
        length = 0
        for item in pattern_dict[abbr]:
            length += item[1] - item[0]

        return length
        

    @staticmethod
    def get_scope_bc(chemistry, wells=""):
        """Return (linker file path, whitelist file path)"""
        try:
            if chemistry == 'bulk_rna':
                linker_f = None
                whitelist_f = f'{ROOT_PATH}/data/chemistry/{chemistry}/bclist{wells}'
            else:
                linker_f = glob.glob(f'{ROOT_PATH}/data/chemistry/{chemistry}/linker*')[0]
                whitelist_f = f'{ROOT_PATH}/data/chemistry/{chemistry}/bclist'
        except IndexError:
            return None, None
        return linker_f, whitelist_f

    @staticmethod
    def ord2chr(q, offset=33):
        return chr(int(q) + offset)

    @staticmethod
    def qual_int(char, offset=33):
        return ord(char) - offset

    @staticmethod
    def low_qual(quals, minQ, num):
        # print(ord('/')-33)           14
        return True if len([q for q in quals if Barcode.qual_int(q) < minQ]) > num else False

    @staticmethod
    def findall_mismatch(seq, n_mismatch=1, bases='ACGTN'):
        """
        choose locations where there's going to be a mismatch using combinations
        and then construct all satisfying lists using product

        Return:
        all mismatch <= n_mismatch set. 

        >>> answer = set(["TCG", "AAG", "ACC", "ATG", "ACT", "ACN", "GCG", "ANG", "ACA", "ACG", "CCG", "AGG", "NCG"])
        >>> seq_set = Barcode.findall_mismatch("ACG")
        >>> seq_set == answer
        True
        """
        seq_set = set()
        seq_len = len(seq)
        if n_mismatch > seq_len:
            n_mismatch = seq_len
        for locs in combinations(range(seq_len), n_mismatch):
            seq_locs = [[base] for base in seq]
            for loc in locs:
                seq_locs[loc] = list(bases)
            for poss in product(*seq_locs):
                seq_set.add(''.join(poss))
        return seq_set

    @staticmethod
    def get_mismatch_dict(seq_list, n_mismatch=1):
        """
        Return:
        mismatch dict. Key: mismatch seq, value: seq in seq_list

        >>> seq_list = ["AACGTGAT", "AAACATCG"]
        >>> mismatch_dict = Barcode.get_mismatch_dict(seq_list)
        >>> mismatch_dict["AACGTGAA"] == "AACGTGAT"
        True
        """
        mismatch_dict = {}

        for seq in seq_list:
            seq = seq.strip()
            if seq == '':
                continue
            for mismatch_seq in Barcode.findall_mismatch(seq, n_mismatch):
                mismatch_dict[mismatch_seq] = seq

        return mismatch_dict

    @staticmethod
    def check_seq_mismatch(seq_list, correct_set_list, mismatch_dict_list):
        '''
        Return bool_valid, bool_corrected, corrected_seq

        >>> seq_list = ['ATA', 'AAT', 'ATA']
        >>> correct_set_list = [{'AAA'},{'AAA'},{'AAA'}]
        >>> mismatch_dict_list = [Barcode.get_mismatch_dict(['AAA'])] * 3

        >>> Barcode.check_seq_mismatch(seq_list, correct_set_list, mismatch_dict_list)
        (True, True, 'AAA_AAA_AAA')

        >>> seq_list = ['AAA', 'AAA', 'AAA']
        >>> Barcode.check_seq_mismatch(seq_list, correct_set_list, mismatch_dict_list)
        (True, False, 'AAA_AAA_AAA')
        '''
        bool_valid = True
        bool_corrected = False
        corrected_seq_list = []
        for index, seq in enumerate(seq_list):
            if seq not in correct_set_list[index]:
                if seq not in mismatch_dict_list[index]:
                    bool_valid = False
                    return bool_valid, bool_corrected, ""
                else:
                    bool_corrected = True
                    corrected_seq_list.append(mismatch_dict_list[index][seq])
            else:
                corrected_seq_list.append(seq)

        return bool_valid, bool_corrected, '_'.join(corrected_seq_list)

    @staticmethod
    def parse_whitelist_file(files: list, n_pattern: int, n_mismatch: int):
        """
        files: file paths
        n_pattern: number of sections in pattern
        n_mismatch: allowed number of mismatch bases
        Returns:
            white_set_list
            mismatch_list
        """
        n_files = len(files)
        if n_files == 1 and n_pattern > 1:
            files = [files[0]] * n_pattern
        elif n_files != n_pattern:
            sys.exit(f'number of whitelist files({n_files} files:{files}) != n_pattern({n_pattern})')
        
        white_set_list, mismatch_list = [], []
        for f in files:
            barcodes, _ = utils.read_one_col(f)
            white_set_list.append(set(barcodes))
            barcode_mismatch_dict = Barcode.get_mismatch_dict(barcodes, n_mismatch)
            mismatch_list.append(barcode_mismatch_dict)

        return white_set_list, mismatch_list

    @staticmethod
    def parse_chemistry(chemistry):
        """
        Returns: pattern_dict, barcode_set_list, barcode_mismatch_list, linker_set_list, linker_mismatch_list
        """
        pattern = PATTERN_DICT[chemistry]
        pattern_dict = Barcode.parse_pattern(pattern)
        linker_file, whitelist_file = Barcode.get_scope_bc(chemistry)

        barcode_set_list, barcode_mismatch_list = Barcode.parse_whitelist_file([whitelist_file], n_pattern=len(pattern_dict['C']), n_mismatch=1)
        linker_set_list, linker_mismatch_list = Barcode.parse_whitelist_file([linker_file],n_pattern=1, n_mismatch=2)

        return pattern_dict, barcode_set_list, barcode_mismatch_list, linker_set_list, linker_mismatch_list

    @staticmethod
    def check_polyT(seq, pattern_dict, min_polyT_count=MIN_T):
        """
        Return:
            True if polyT is found
        """
        seq_polyT = Barcode.get_seq_str(seq, pattern_dict['T'])
        n_polyT_found = seq_polyT.count('T')
        if n_polyT_found >= min_polyT_count:
            return True
        return False

    def open_files(self):
        if self.output_R1 or self.bool_flv:
            self.fh_fq1 = xopen(self.out_fq1, 'w')
        if not self.args.stdout:
            self.fh_fq2 = xopen(self.out_fq2, 'w')

        if self.nopolyT:
            self.fh_nopolyT_fq1 = xopen(self.nopolyT_1, 'w')
            self.fh_nopolyT_fq2 = xopen(self.nopolyT_2, 'w')

        if self.noLinker:
            self.fh_nolinker_fq1 = xopen(self.noLinker_1, 'w')
            self.fh_nolinker_fq2 = xopen(self.noLinker_2, 'w')

    def close_files(self):
        if self.output_R1 or self.bool_flv:
            self.fh_fq1.close()
        if not self.args.stdout:
            self.fh_fq2.close()

        if self.nopolyT:
            self.fh_nopolyT_fq1.close()
            self.fh_nopolyT_fq2.close()
        
        if self.noLinker:
            self.fh_nolinker_fq1.close()
            self.fh_nolinker_fq2.close()

    @utils.add_log
    def add_step_metrics(self):

        self.add_metric(
            name='Raw Reads',
            value=self.total_num,
            help_info='total reads from FASTQ files'
        )
        self.add_metric(
            name='Valid Reads',
            value=self.clean_num,
            total=self.total_num,
            help_info='reads pass filtering(filtered: reads without poly T, reads without linker, reads without correct barcode or low quality reads)'
        )

        BarcodesQ30 = sum([self.barcode_qual_Counter[k] for k in self.barcode_qual_Counter if k >= self.ord2chr(
            30)]) / float(sum(self.barcode_qual_Counter.values())) * 100
        BarcodesQ30 = round(BarcodesQ30, 2)
        BarcodesQ30_display = f'{BarcodesQ30}%'
        self.add_metric(
            name='Q30 of Barcodes',
            value=BarcodesQ30,
            display=BarcodesQ30_display,
            help_info='percent of barcode base pairs with quality scores over Q30',
        )

        UMIsQ30 = sum([self.umi_qual_Counter[k] for k in self.umi_qual_Counter if k >= self.ord2chr(
            30)]) / float(sum(self.umi_qual_Counter.values())) * 100
        UMIsQ30 = round(UMIsQ30, 2)
        UMIsQ30_display = f'{UMIsQ30}%'
        self.add_metric(
            name='Q30 of UMIs',
            value=UMIsQ30,
            display=UMIsQ30_display,
            help_info='percent of UMI base pairs with quality scores over Q30',
        )

        self.add_metric(
            name='No PolyT Reads',
            value=self.no_polyT_num,
            total=self.total_num,
            show=False
        )

        self.add_metric(
            name='Low Quality Reads',
            value=self.lowQual_num,
            total=self.total_num,
            show=False,
        )

        self.add_metric(
            name='No Linker Reads',
            value=self.no_linker_num,
            total=self.total_num,
            show=False,
        )

        self.add_metric(
            name='No Barcode Reads',
            value=self.no_barcode_num,
            total=self.total_num,
            show=False,
        )

        self.add_metric(
            name='Corrected Linker Reads',
            value=self.linker_corrected_num,
            total=self.total_num,
            show=False,
        )

        self.add_metric(
            name='Corrected Barcode Reads',
            value=self.barcode_corrected_num,
            total=self.total_num,
            show=False,
        )

        if self.clean_num == 0:
            raise Exception('no valid reads found! please check the --chemistry parameter.' + HELP_DICT['chemistry'])
        
        if self.assay == 'flv_trust4':
            self.add_metric(
                name='Valid Matched Reads',
                value=self.match_num,
                total=self.total_num,
                help_info='reads match with flv_rna cell barcodes'
            )

            self.add_metric(
                name='Matched Barcodes',
                value=len(self.match_cbs),
                help_info='barcodes match with flv_rna'
            )


    @utils.add_log
    def run(self):
        """
        Extract barcode and UMI from R1. Filter reads with 
            - invalid polyT
            - low quality in barcode and UMI
            - invalid inlinker
            - invalid barcode
            
        for every sample
            get chemistry
            get linker_mismatch_dict and barcode_mismatch_dict
            for every read in read1
                filter
                write valid R2 read to file
        """

        for i in range(self.fq_number):

            chemistry = self.chemistry_list[i]
            lowNum = int(self.lowNum)
            lowQual = int(self.lowQual)
            if chemistry == 'scopeV1':
                lowNum = min(0, lowNum)
                lowQual = max(10, lowQual)
                self.run.logger.info(f'scopeV1: lowNum={lowNum}, lowQual={lowQual} ')
            # get linker and whitelist
            bc_pattern = PATTERN_DICT[chemistry]
            if (bc_pattern):
                linker_file, whitelist_file = self.get_scope_bc(chemistry, self.wells)
                whitelist_files = [whitelist_file]
            else:
                bc_pattern = self.pattern
                linker_file = self.linker
                whitelist_file = self.whitelist
                whitelist_files = whitelist_file.split(',')
            if not bc_pattern:
                raise Exception("invalid bc_pattern!")

            pattern_dict = self.parse_pattern(bc_pattern)

            bool_T = True if 'T' in pattern_dict else False
            bool_L = True if 'L' in pattern_dict else False
            bool_whitelist = (whitelist_file is not None) and whitelist_file != "None"
            C_len = sum([item[1] - item[0] for item in pattern_dict['C']])

            if bool_whitelist:
                barcode_set_list, barcode_mismatch_list = Barcode.parse_whitelist_file(whitelist_files,
                                                n_pattern=len(pattern_dict['C']), n_mismatch=1)
            if bool_L:
                linker_set_list, linker_mismatch_list = Barcode.parse_whitelist_file([linker_file],n_pattern=1, n_mismatch=2)

            with pysam.FastxFile(self.fq1_list[i], persist=False) as fq1, \
                    pysam.FastxFile(self.fq2_list[i], persist=False) as fq2:
                for entry1, entry2 in zip(fq1, fq2):
                    header1, seq1, qual1 = entry1.name, entry1.sequence, entry1.quality
                    header2, seq2, qual2 = entry2.name, entry2.sequence, entry2.quality
                    self.total_num += 1

                    # polyT filter
                    if bool_T and self.filterNoPolyT:
                        if not Barcode.check_polyT(seq1, pattern_dict):
                            self.no_polyT_num += 1
                            if self.nopolyT:
                                self.fh_nopolyT_fq1.write(
                                    '@%s\n%s\n+\n%s\n' % (header1, seq1, qual1))
                                self.fh_nopolyT_fq2.write(
                                    '@%s\n%s\n+\n%s\n' % (header2, seq2, qual2))
                            continue

                    # lowQual filter
                    C_U_quals_ascii = Barcode.get_seq_str(
                        qual1, pattern_dict['C'] + pattern_dict['U'])
                    if lowQual > 0 and Barcode.low_qual(C_U_quals_ascii, lowQual, lowNum):
                        self.lowQual_num += 1
                        continue

                    # linker filter
                    if bool_L and (not self.allowNoLinker):
                        seq_str = Barcode.get_seq_str(seq1, pattern_dict['L'])
                        bool_valid, bool_corrected, _ = Barcode.check_seq_mismatch(
                            [seq_str], linker_set_list, linker_mismatch_list)
                        if not bool_valid:
                            self.no_linker_num += 1
                            if self.noLinker:
                                self.fh_nolinker_fq1.write(
                                    f'@{header1}\n{seq1}\n{seq_str}\n{qual1}\n')
                                self.fh_nolinker_fq2.write(
                                    '@%s\n%s\n+\n%s\n' % (header2, seq2, qual2))
                            continue
                        elif bool_corrected:
                            self.linker_corrected_num += 1

                    # barcode filter
                    seq_list = self.get_seq_list(seq1, pattern_dict, 'C')
                    if self.bool_flv:
                        seq_list = [utils.reverse_complement(seq) for seq in seq_list[::-1]]
                    if bool_whitelist:
                        bool_valid, bool_corrected, corrected_seq = Barcode.check_seq_mismatch(
                            seq_list, barcode_set_list, barcode_mismatch_list)

                        if not bool_valid:
                            self.no_barcode_num += 1
                            continue
                        elif bool_corrected:
                            self.barcode_corrected_num += 1
                        cb = corrected_seq
                    else:
                        cb = "_".join(seq_list)

                    self.clean_num += 1
                    self.barcode_qual_Counter.update(C_U_quals_ascii[:C_len])
                    self.umi_qual_Counter.update(C_U_quals_ascii[C_len:])

                    umi = Barcode.get_seq_str(seq1, pattern_dict['U'])
                    if not umi:
                        continue

                    if self.bool_flv:
                        qual1 = 'F' * len(cb + umi)
                        self.barcode_read_Counter.update(cb)
                        if self.assay == 'flv_trust4' and cb in self.match_barcodes:
                            self.match_num += 1
                            self.match_cbs.add(cb)
                            if self.barcode_read_Counter[cb] <= 80000:
                                self.fh_fq2.write(f'@{cb}:{umi}:{self.total_num}\n{seq2}\n+\n{qual2}\n')
                                self.fh_fq1.write(f'@{cb}:{umi}:{self.total_num}\n{cb}{umi}\n+\n{qual1}\n')
                    else:
                        if self.args.stdout:
                            print(f'@{cb}:{umi}:{self.total_num}\n{seq2}\n+\n{qual2}')
                        else:
                            self.fh_fq2.write(f'@{cb}:{umi}:{self.total_num}\n{seq2}\n+\n{qual2}\n')
                        if self.output_R1:
                            self.fh_fq1.write(f'@{cb}:{umi}:{self.total_num}\n{seq1}\n+\n{qual1}\n')                   
            
                self.run.logger.info(self.fq1_list[i] + ' finished.')

        self.close_files()
        self.add_step_metrics()


@utils.add_log
def barcode(args):
    with Barcode(args, display_title='Demultiplexing') as runner:
        runner.run()


def get_opts_barcode(parser, sub_program=True):
    parser.add_argument(
        '--chemistry',
        help='Predefined (pattern, barcode whitelist, linker whitelist) combinations. ' + HELP_DICT['chemistry'],
        choices=list(PATTERN_DICT.keys()),
        default='auto'
    )
    parser.add_argument(
        '--pattern',
        help="""The pattern of R1 reads, e.g. `C8L16C8L16C8L1U12T18`. The number after the letter represents the number 
        of bases.  
- `C`: cell barcode  
- `L`: linker(common sequences)  
- `U`: UMI    
- `T`: poly T""",
    )
    parser.add_argument(
        '--whitelist',
        help='Cell barcode whitelist file path, one cell barcode per line.'
    )
    parser.add_argument(
        '--linker',
        help='Linker whitelist file path, one linker per line.'
    )
    parser.add_argument(
        '--lowQual',
        help='Default 0. Bases in cell barcode and UMI whose phred value are lower than \
lowQual will be regarded as low-quality bases.',
        type=int,
        default=0
    )
    parser.add_argument(
        '--lowNum',
        help='The maximum allowed lowQual bases in cell barcode and UMI.',
        type=int,
        default=2
    )
    parser.add_argument(
        '--nopolyT',
        help='Outputs R1 reads without polyT.',
        action='store_true',
    )
    parser.add_argument(
        '--noLinker',
        help='Outputs R1 reads without correct linker.',
        action='store_true',
    )
    parser.add_argument(
        '--filterNoPolyT',
        help="Filter reads without PolyT.",
        action='store_true'
    )
    parser.add_argument(
        '--allowNoLinker',
        help="Allow valid reads without correct linker.",
        action='store_true'
    )
    parser.add_argument(
        '--gzip',
        help="Output gzipped fastq files.",
        action='store_true'
    )
    parser.add_argument(
        '--output_R1',
        help="Output valid R1 reads.",
        action='store_true'
    )
    parser.add_argument(
        '--wells',
        help='The AccuraCode wells used (384 or 96).',
        type=int,
        default=384
    )
    if sub_program:
        parser.add_argument('--fq1', help='R1 fastq file. Multiple files are separated by comma.', required=True)
        parser.add_argument('--fq2', help='R2 fastq file. Multiple files are separated by comma.', required=True)
        parser.add_argument('--match_dir', help='Matched scRNA-seq directory, required for flv_trust4')
        parser.add_argument(
            '--stdout',
            help="Write output to standard output",
            action='store_true'
    )
        parser = s_common(parser)

    return parser


if __name__ == '__main__':
    unittest.main()



import abc
import sys
import io
import json
import numbers
import os
import subprocess

from jinja2 import Environment, FileSystemLoader, select_autoescape

from celescope.tools import utils
from celescope.__init__ import HELP_DICT, __version__


def cap_str_except_preposition(my_string):
    prepositions = {"and", "or", "the", "a", "of", "in", "per", "after", 'with', 'by', 'at'}
    lowercase_words = my_string.split(" ")

    final_words = [word if word in prepositions else word[0].upper() + word[1:] for word in lowercase_words]
    final_words = " ".join(final_words)
    return final_words


def s_common(parser):
    """subparser common arguments
    """
    parser.add_argument('--outdir', help='Output diretory.', required=True)
    parser.add_argument('--sample', help='Sample name.', required=True)
    parser.add_argument('--thread', help=HELP_DICT['thread'], default=4)
    parser.add_argument('--debug', help=HELP_DICT['debug'], action='store_true')
    return parser


class Step:
    """
    Step class
    """

    def __init__(self, args, display_title=None):
        '''
        display_title controls the section title in HTML report
        force thread <=20
        '''
        sys.stderr.write(f'CeleScope version: {__version__} ')
        sys.stderr.write(f'Args: {args}\n')
        self.args = args
        self.outdir = args.outdir
        self.outs_dir = f'{args.outdir}/../outs'
        self.sample = args.sample
        self.assay = args.subparser_assay
        self.thread = int(args.thread)
        self.thread = min(self.thread, 20)
        self.debug = args.debug
        self.out_prefix = f'{self.outdir}/{self.sample}'
        self.display_title = display_title

        # metrics index
        self.metric_index = 0

        # important! make outdir before path_dict because path_dict use relative path.
        utils.check_mkdir(self.outdir)
        utils.check_mkdir(self.outs_dir)

        # set
        class_name = self.__class__.__name__
        if not display_title:
            self._display_title = class_name
        else:
            self._display_title = display_title
        self._step_name = class_name[0].lower() + class_name[1:]
        self.__slots = ['data', 'metrics']
        self._step_summary_name = f'{self._step_name}_summary'

        self.__metric_list = []
        self.__help_content = []
        self.__comments = []
        self._path_dict = {}
        for slot in self.__slots:
            self._path_dict[slot] = f'{self.outdir}/../.{slot}.json'

        self.__content_dict = {}
        self.old_step_dict = {}
        for slot, path in self._path_dict.items():
            if not os.path.exists(path):
                self.__content_dict[slot] = {}
            else:
                with open(path) as f:
                    try:
                        self.__content_dict[slot] = json.load(f)
                    except ValueError:
                        print(f'WARNING: Decoding "{path}" as json has failed. Will create empty json file.')
                        self.__content_dict[slot] = {}
            # clear step_summary
            if self._step_summary_name in self.__content_dict[slot]:
                self.old_step_dict[slot] = self.__content_dict[slot][self._step_summary_name]
            self.__content_dict[slot][self._step_summary_name] = {}

        # jinja env
        self.env = Environment(
            loader=FileSystemLoader(os.path.dirname(__file__) + '/../templates/'),
            autoescape=select_autoescape(['html', 'xml'])
        )

        # out file
        self.__stat_file = f'{self.outdir}/stat.txt'
        self.report_html = f"{self.outdir}/../{self.sample}_report.html"

        # move file to outs
        self.outs = []

    def add_metric(self, name, value, total=None, help_info=None, display=None, show=True, print_log=True, value_type=None):
        '''
        add metric to metric_list
        
        Args
            total: int or float, used to calculate fraction
            help_info: str, help info for metric in html report
            display: str, controls how to display the metric in HTML report.
            show: bool, whether to add to `.data.json` and `stat.txt`. `.data.json` is used for HTML report. `stat.txt` is used in house.
            print_log: bool, whether to print metric to stdout
            value_type: if value_type is "fraction", value = round(value * 100, 2), display = f'{value}%'
        '''

        name = cap_str_except_preposition(name)
        if help_info:
            help_info = help_info[0].upper() + help_info[1:]
            if help_info[-1] != '.':
                help_info += '.'
        if not display:
            if isinstance(value, numbers.Number):
                display = str(format(value, ','))
            else:
                display = value
        fraction = None
        if total:
            fraction = round(value / total * 100, 2)
            display += f'({fraction}%)'
        if value_type == 'fraction':
            value = round(value * 100, 2)
            display = f'{value}%'
        self.__metric_list.append(
            {
                "name": name,
                "value": value,
                "total": total,
                "fraction": fraction,
                "display": display,
                "help_info": help_info,
                "show": show,
            }
        )

        if print_log:
            sys.stderr.write(f'{name}: {display}\n')

    def _write_stat(self):
        with open(self.__stat_file, 'w') as writer:
            for metric in self.__metric_list:
                if metric['show']:
                    name = metric['name']
                    display = metric['display']

                    line = f'{name}: {display}'
                    writer.write(line + '\n')

    def _dump_content(self):
        '''dump content to json file
        '''
        for slot, path in self._path_dict.items():
            if self.__content_dict[slot]:
                with open(path, 'w') as f:
                    json.dump(self.__content_dict[slot], f, indent=4)

    @utils.add_log
    def _render_html(self):
        template = self.env.get_template(f'html/{self.assay}/base.html')
        with io.open(self.report_html, 'w', encoding='utf8') as f:
            html = template.render(self.__content_dict['data'])
            f.write(html)

    def _add_content_data(self):
        step_summary = {}
        step_summary['display_title'] = self._display_title
        metric_list = []
        comment_metric_list = []
        for metric in self.__metric_list:
            if metric['show']:
                metric_list.append(metric)
            else:
                comment_metric_list.append(metric)
        step_summary['metric_list'] = metric_list
        step_summary['comment_metric_list'] = comment_metric_list
        step_summary['help_content'] = self.__help_content
        step_summary['comments'] = self.__comments
        self.__content_dict['data'][self._step_summary_name].update(step_summary)

    def _add_content_metric(self):
        metric_dict = dict()
        for metric in self.__metric_list:
            name = metric['name']
            value = metric['value']
            fraction = metric['fraction']
            metric_dict[name] = value
            if fraction:
                metric_dict[f'{name} Fraction'] = fraction

        self.__content_dict['metrics'][self._step_summary_name].update(metric_dict)

    def add_data(self, **kwargs):
        """
        add data(other than metrics) to self.content_dict['data']
        for example: add plots and tables
        """
        for key, value in kwargs.items():
            self.__content_dict['data'][self._step_summary_name][key] = value

    def add_help_content(self, name, content):
        """
        add help info before metrics' help_info
        """
        self.__help_content.append(
            {
                'name': name,
                'content': content
            }
        )

    def add_comments(self, content):
        """
        add comment to self.__comments
        """
        self.__comments.append(content)

    @utils.add_log
    def get_slot_key(self, slot, step_name, key):
        '''read slot from json file
        '''
        try:
            return self.__content_dict[slot][step_name + '_summary'][key]
        except KeyError:
            self.get_slot_key.logger.warning(f'{key} not found in {step_name}_summary.{slot}')
            raise

    @utils.add_log
    def add_slot_step(self, slot, step_name, val):
        '''add slot to json
        '''
        self.__content_dict[slot][step_name + '_summary'] = val

    @utils.add_log
    def get_slot_step(self, slot, step_name):
        '''add slot to json
        '''
        return self.__content_dict[slot][step_name + '_summary']



    def get_table_dict(self, title, table_id, df_table):
        """
        table_dict {title: '', table_id: '', df_table: pd.DataFrame}
        """
        table_dict = {}
        table_dict['title'] = title
        table_dict['table'] = df_table.to_html(
            escape=False,
            index=False,
            table_id=table_id,
            justify="center")
        table_dict['id'] = table_id
        return table_dict
    
    def _move_files(self):
        for f in self.outs:
            cmd = ''
            if not os.path.exists(f):
                sys.stderr.write(f'WARNING: output file {f} not found! The pipeline may have failed.\n')
            elif os.path.isfile(f):
                cmd = f'mv -f {f} {self.outs_dir}'
            elif os.path.isdir(f):
                cmd = f'set -e; cp -r {f} {self.outs_dir}; rm -r {f}'
            subprocess.check_call(cmd, shell=True)


    @utils.add_log
    def _clean_up(self):
        self._add_content_data()
        self._add_content_metric()
        self._write_stat()
        self._dump_content()
        self._render_html()
        self._move_files()

    @utils.add_log
    def debug_subprocess_call(self, cmd):
        '''
        debug subprocess call
        '''
        self.debug_subprocess_call.logger.info(cmd)
        if cmd.find('2>&1') == -1:
            cmd += ' 2>&1 '
        subprocess.check_call(cmd, shell=True)

    def get_metric_list(self):
        return self.__metric_list

    def set_metric_list(self, metric_list):
        self.__metric_list = metric_list

    @abc.abstractmethod
    def run(self):
        sys.exit('Please implement run() method.')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self._clean_up()



from celescope.dynaseq.__init__ import __ASSAY__
from celescope.rna.multi_rna import Multi_rna
from celescope.tools.__init__ import FILTERED_MATRIX_DIR_SUFFIX, BARCODE_FILE_NAME, STARSOLO_BAM_SUFFIX, OUTS_DIR


class Multi_dynaseq(Multi_rna):

    """
    ## Usage

    ```
        multi_dynaseq\\
        --mapfile ./rna.mapfile\\
        --genomeDir /SGRNJ/Public/Database/genome/homo_mus\\
    ```

    For control sample, set --control to skip replacement step.
    ```
        multi_dynaseq\\
        --mapfile ./rna.mapfile\\
        --genomeDir /SGRNJ/Public/Database/genome/homo_mus\\
        --control
    ```
    """
    def starsolo(self, sample):
        step = 'starsolo'
        arr = self.fq_dict[sample]
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq1 {arr[0]} --fq2 {arr[1]} '
            f'--STAR_param "--outFilterScoreMinOverLread 0.3 --outFilterMatchNminOverLread 0.3" '
            f'--SAM_attributes MD '
        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.thread)

    def conversion(self, sample):
        step = 'conversion'
        bam = f'{self.outdir_dic[sample]["outs"]}/{sample}_{STARSOLO_BAM_SUFFIX}'
        cell = f'{self.outdir_dic[sample]["outs"]}/{FILTERED_MATRIX_DIR_SUFFIX}/{BARCODE_FILE_NAME}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
            f'--cell {cell} '
        )
        self.process_cmd(cmd, step, sample, m=self.args.conversionMem, x=self.args.thread)

    def substitution(self, sample):
        step = 'substitution'
        bam = f'{self.outdir_dic[sample]["conversion"]}/{sample}.PosTag.bam'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
        )
        self.process_cmd(cmd, step, sample, m=1, x=1)

    def replacement(self, sample):
        step = 'replacement'
        bam = f'{self.outdir_dic[sample]["conversion"]}/{sample}.PosTag.bam'
        snp = f'{self.outdir_dic[sample]["conversion"]}/{sample}.snp.csv'
        tsne_file = f'{self.outdir_dic[sample][OUTS_DIR]}/tsne_coord.tsv'
        cell = f'{self.outdir_dic[sample]["outs"]}/{FILTERED_MATRIX_DIR_SUFFIX}/{BARCODE_FILE_NAME}'
        cmd_line = self.get_cmd_line(step, sample)
        bg_para = ''
        if sample in self.col5_dict:
            bg_para = f'--bg {self.col5_dict[sample]} '
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
            f'--bg {snp} {bg_para} '
            f'--tsne {tsne_file} '
            f'--cell {cell} '
        )
        self.process_cmd(cmd, step, sample, m=5*int(self.args.thread), x=self.args.thread)



def main():
    multi = Multi_dynaseq(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()


"""
Add zeor count genes to matrix
"""
import argparse
import os

import scipy

from celescope.tools.matrix import CountMatrix, Features
from celescope.tools.reference import GtfParser
from celescope.tools import utils


def add_zero_count(cm: CountMatrix, all_features: Features):
    """
    Returns:
        CountMatrix object
    """
    m = cm.get_matrix()

    all_gene_id = all_features.gene_id
    feature_index_dict = {} # {'ENSG00000223972': 0, 'ENSG00000227232': 1,}
    for index, gene_id in enumerate(all_gene_id):
        feature_index_dict[gene_id] = index

    row_index_dict = {} # {0: 25969, 1: 54488}
    for sub_index, gene_id in enumerate(cm.get_features().gene_id):
        row_index_dict[sub_index] = feature_index_dict[gene_id]

    new_row = [row_index_dict[sub_index] for sub_index in m.row]
    new_matrix = scipy.sparse.coo_matrix((m.data, (new_row, m.col)), 
            shape=(len(feature_index_dict), len(cm.get_barcodes())))

    return CountMatrix(all_features, cm.get_barcodes(), new_matrix)



if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Add zeor count genes to matrix')
    parser.add_argument('--matrix_list_file', help='Plain text file with one matrix dir per line.', required=True)
    parser.add_argument('--gtf', required=True, help='GTF file.')
    parser.add_argument('--outdir', help='Output directory.', default='./out')

    args = parser.parse_args()

    os.makedirs(args.outdir, exist_ok=True)

    gp = GtfParser(args.gtf)
    gp.get_id_name()
    all_features = gp.get_features()

    matrix_list, _ = utils.read_one_col(args.matrix_list_file)
    for matrix_dir in matrix_list:
        matrix_name = matrix_dir.split('/')[-1]
        out_matrix_dir = f'{args.outdir}/{matrix_name}'
        cm = CountMatrix.from_matrix_dir(matrix_dir)
        new_cm = add_zero_count(cm, all_features)
        new_cm.to_matrix_dir(out_matrix_dir)




from celescope.tools import utils
from celescope.tools import reference



class Mkgtf:
    """
    ##Features

    - Filter GTF files.
    
    GTF files downloaded from sites like ENSEMBL and UCSC often contain transcripts and genes which need to be filtered.
    Celescope provides mkgtf, a simple utility to filter genes. The command syntax requires input and output fitlered GTF file names.

    The following gene biotypes will be kept.
    ```
    protein_coding 
    lncRNA 
    antisense 
    IG_LV_gene 
    IG_V_gene 
    IG_V_pseudogene 
    IG_D_gene 
    IG_J_gene 
    IG_J_pseudogene 
    IG_C_gene 
    IG_C_pseudogene 
    TR_V_gene 
    TR_V_pseudogene 
    TR_D_gene 
    TR_J_gene 
    TR_J_pseudogene 
    TR_C_gene
    ```
    """

    def __init__(self, in_gtf, out_gtf, attributes):


        self.in_gtf_fn = in_gtf
        self.out_gtf_fn = out_gtf

        self.attributes = {}
        for attr_str in attributes.split(';'):
            if attr_str:
                attr, val = attr_str.split('=')
                val = set(val.split(','))
                self.attributes[attr] = val

    @utils.add_log
    def run(self):
        runner = reference.GtfBuilder(self.in_gtf_fn, self.out_gtf_fn, self.attributes)
        runner.build_gtf()


def mkgtf(args):
    runner = Mkgtf(args.gtf, args.out_gtf, args.attributes)
    runner.run()


def get_opts_mkgtf(parser, sub_program=True):
    if sub_program:
        parser.add_argument('gtf', help='raw gtf file')
        parser.add_argument('out_gtf', help='output gtf file')
        parser.add_argument('--add_intron', help='add intron lines to gtf file', action='store_true')
        parser.add_argument(
            '--attributes', 
            help='Attributes to keep. Example: `gene_biotype=protein_coding,lncRNA,antisense;`',
            default="gene_biotype=protein_coding,lncRNA,antisense,IG_LV_gene,IG_V_gene,IG_V_pseudogene,IG_D_gene,IG_J_gene,IG_J_pseudogene,IG_C_gene,IG_C_pseudogene,TR_V_gene,TR_V_pseudogene,TR_D_gene,TR_J_gene,TR_J_pseudogene,TR_C_gene;",
            )

    return parser


import argparse
import json
import os
from collections import defaultdict

import pandas as pd
from celescope.tools import utils


@utils.add_log
def append_sample_data(sample, sample_data_dict, all_data_dict, steps):
    """
    append sample data to all_data_dict
    Args:
        all_data_dict - key: step, value: list of sample series
    """
    for step in steps:
        title = step + '_summary'
        if title in sample_data_dict:
            metric_list = sample_data_dict[title]['metric_list']
            if metric_list:
                metric_dict = {'sample': sample}
                for metric in metric_list:
                    name = metric['name'].replace(' ', '_')
                    metric_dict[name] = metric['display']

                all_data_dict[step].append(pd.Series(metric_dict))

@utils.add_log
def write_merge_report(all_data_dict, merge_report_handle, steps):
    """
    merge_report:
        ## sample_summary
        sample,sample_ID,Assay...
        sample_1,sample_1,rna...
        ## barcode_summary

    Args:
        all_data_dict - key: title, value: list of sample series
    """

    for step in steps:
        if all_data_dict[step]:
            merge_report_handle.write(f'## {step}_summary\n')
            df = pd.concat(all_data_dict[step], axis=1).T
            df.to_csv(merge_report_handle, index=False, mode='a', sep='\t')
            merge_report_handle.write('\n')

@utils.add_log
def run(args):
    # out
    os.chdir(args.outdir)
    out_file = 'merge.xls'

    all_data_dict = defaultdict(list) 
    steps = args.steps.split(',')
    samples = args.samples.split(',')
    run.logger.info(f'samples: {samples}')

    for sample in samples:
        sample_data_dict = json.load(open(f'{sample}/.data.json', 'r'))
        append_sample_data(sample, sample_data_dict, all_data_dict, steps)

    write_merge_report(all_data_dict, open(out_file, 'w'), steps)


def main():
    parser = argparse.ArgumentParser('merge report')
    parser.add_argument('--outdir', help='outdir', required=True)
    parser.add_argument('--samples', help='samples, seperated by comma', required=True)
    parser.add_argument('--steps', help='steps', required=True)
    parser.add_argument('--rm_files', action='store_true', help='remove all fq and bam after running')
    args = parser.parse_args()

    run(args)

@utils.add_log
def rm_files():
    cmd = '''
        find . -iname '*.fq*' -delete;
        find . -iname '*.bam' -not -path './*/*.featureCounts/*name_sorted.bam' -delete;
    '''
    os.system(cmd)


if __name__ == '__main__':
    main()



import unittest
import os
from collections import namedtuple
from celescope.snp.variant_calling import Variant_calling

ROOT_DIR = os.path.dirname(__file__)


class Test_variant_calling(unittest.TestCase):
    def setUp(self):
        os.chdir(ROOT_DIR)
        Args = namedtuple("Args", "thread outdir sample assay debug " + "genomeDir vcf bam match_dir")
        self.args = Args(
            thread=10,
            outdir="./test_output/07.variant_calling",
            sample="test1",
            assay="snp",
            debug=False,
            genomeDir="/SGRNJ/Public/Database/genome/homo_sapiens/ensembl_92",
            vcf=None,
            bam="./test_data/06.target_metrics/subset_filter.bam",
            match_dir="./test_data/match_dir",
        )

    def test_run(self):
        obj = Variant_calling(self.args, "variant_calling")
        '''
        obj.SplitNCigarReads()
        obj.split_bam()
        obj.call_all_snp()
        if obj.vcf_bool:
            obj.add_VID()
        else:
            obj.merge_vcf()
        '''
        obj.write_VID_file()
        obj.get_UMI()
        obj.write_support_matrix()
        obj._clean_up()




"""
split scRNA-Seq fastq file(01.barcode/{sample}_2.fq)
"""
import glob
import os
import itertools
from collections import defaultdict
import sys

import pysam
import pandas as pd

from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.tools.__init__ import FILTERED_MATRIX_DIR_SUFFIX
from celescope.__init__ import HELP_DICT
from celescope.tools.matrix import CountMatrix
from celescope.flv_trust4.annotation import gen_vj_annotation_metrics, gen_clonotypes_table


def get_clonotypes_table(df):
    chains = sorted(set(df['chain'].tolist()))
    res = pd.DataFrame(columns=['barcode'])
    for c in chains:
        df_c = df[df['chain'] == c][['barcode', 'aaSeqCDR3', 'nSeqCDR3']]
        df_c = df_c.rename(columns={'aaSeqCDR3': f'{c}_aaSeqCDR3', 'nSeqCDR3': f'{c}_nSeqCDR3'})
        res = pd.merge(res, df_c, on='barcode', how='outer').fillna('NaN')
    group_l = res.columns.tolist()
    group_l.remove('barcode')
    clonotypes = res.groupby(group_l, as_index=False).agg({'barcode': 'count'})
    clonotypes = clonotypes.rename(columns={'barcode': 'barcode_count'})
    sum_cb = clonotypes['barcode_count'].sum()
    clonotypes['percent'] = clonotypes['barcode_count'].apply(lambda x: round(x/sum_cb, 2))
    clonotypes['clonetype_ID'] = [i for i in range(1, clonotypes.shape[0]+1)]
    group_l.insert(0, 'clonetype_ID')
    group_l.append('barcode_count')
    group_l.append('percent')
    clonotypes = clonotypes.reindex(columns=group_l)
    if chains[0].startswith("TR"):
        return clonotypes, 'TCR'
    elif chains[0].startswith("IG"):
        return clonotypes, 'BCR'


class Split_tag(Step):
    """
    ## Features
    - Split scRNA-Seq fastq according to tag assignment.

    ## Output
    - `matrix/` Matrix files of each tag.(Optional)
    - `fastq/` Fastq files of each tag.(Optional)
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)
        if not (args.split_matrix or args.split_bam or args.split_fastq or args.split_vdj or args.split_fl_vdj):
            return

        # set
        df_umi_tag = pd.read_csv(args.umi_tag_file, sep='\t', index_col=0)
        df_umi_tag = df_umi_tag.rename_axis('barcode').reset_index()
        self.tag_barcode_dict = {tag: set(row["barcode"].tolist()) for tag, row in df_umi_tag.groupby("tag")}

        if args.split_matrix:
            self.matrix_outdir = f'{args.outdir}/matrix/'
            if utils.check_arg_not_none(args, 'match_dir'):
                matrix_dir = utils.get_matrix_dir_from_match_dir(args.match_dir)
            elif utils.check_arg_not_none(args, 'matrix_dir'):
                matrix_dir = args.matrix_dir
            else:
                raise ValueError("--match_dir or --matrix_dir is required.")
            self.count_matrix = CountMatrix.from_matrix_dir(matrix_dir)

        if args.split_bam:
            if not utils.check_arg_not_none(args, 'bam_file'):
                sys.exit('must provide --bam_file to use --split_bam')
            self.bam_outdir = f'{args.outdir}/bam/'
            utils.check_mkdir(self.bam_outdir)

        if args.split_fastq:
            self.rna_fq_file = glob.glob(f'{args.match_dir}/*barcode/*_2.fq*')[0]

            fastq_outdir = f'{args.outdir}/fastqs/'
            os.system(f'mkdir -p {fastq_outdir}')

            self.r2_fastq_files_handle = {}
            self.r1_fastq_files_handle = {}
            for tag in self.tag_barcode_dict:
                r2_fastq_file_name = f'{fastq_outdir}/{tag}_2.fq'
                self.r2_fastq_files_handle[tag] = open(r2_fastq_file_name, 'w')
                r1_fastq_file_name = f'{fastq_outdir}/{tag}_1.fq'
                self.r1_fastq_files_handle[tag] = open(r1_fastq_file_name, 'w')

            self.tag_read_index_dict = defaultdict(set)

        if args.split_vdj:
            self.cell_confident_vdj = glob.glob(f'{args.vdj_dir}/*count_vdj/*cell_confident.tsv*')[0]

            self.vdj_outdir = f'{args.outdir}/vdj/'
            if not os.path.exists(self.vdj_outdir):
                os.system(f'mkdir -p {self.vdj_outdir}')

        if args.split_fl_vdj:
            self.fl_vdj_outdir = f'{args.outdir}/fl_vdj/'
            if not os.path.exists(self.fl_vdj_outdir):
                os.system(f'mkdir -p {self.fl_vdj_outdir}')
            
            self.anno_file = glob.glob(f'{args.vdj_dir}/04.summarize/*_filtered_contig.csv')[0]
            self.fasta_file = glob.glob(f'{args.vdj_dir}/04.summarize/*_filtered_contig.fasta')[0]

    @utils.add_log
    def write_r2_fastq_files(self):
        read_num = 0
        with pysam.FastxFile(self.rna_fq_file, 'r') as rna_fq:
            for read in rna_fq:
                read_num += 1
                attr = read.name.strip("@").split(':')
                barcode = attr[0]
                read_index = int(attr[2])
                for tag in self.tag_barcode_dict:
                    if barcode in self.tag_barcode_dict[tag]:
                        self.tag_read_index_dict[tag].add(read_index)
                        self.r2_fastq_files_handle[tag].write(str(read) + '\n')

                if read_num % 1000000 == 0:
                    self.write_r2_fastq_files.logger.info(f'{read_num} done')

        for tag in self.r2_fastq_files_handle:
            self.r2_fastq_files_handle[tag].close()

    @utils.add_log
    def write_r1_fastq_files(self):
        file_handles = [pysam.FastxFile(r1, 'r') for r1 in self.args.R1_read.split(',')]
        r1_read = itertools.chain(*file_handles)
        for read_index, read in enumerate(r1_read, start=1):
            for tag in self.tag_read_index_dict:
                if read_index in self.tag_read_index_dict[tag]:
                    self.r1_fastq_files_handle[tag].write(str(read) + '\n')

        for r1 in file_handles:
            r1.close()
        for tag in self.r1_fastq_files_handle:
            self.r1_fastq_files_handle[tag].close()

    @utils.add_log
    def split_matrix(self):
        for tag in self.tag_barcode_dict:
            tag_barcodes = list(self.tag_barcode_dict[tag])
            raw_barcodes = self.count_matrix.get_barcodes()
            tag_barcodes_indices = [raw_barcodes.index(barcode) for barcode in tag_barcodes]
            tag_barcodes_indices.sort()
            slice_matrix = self.count_matrix.slice_matrix(tag_barcodes_indices)

            tag_matrix_dir = f'{self.matrix_outdir}/{tag}_{FILTERED_MATRIX_DIR_SUFFIX}/'
            slice_matrix.to_matrix_dir(tag_matrix_dir)

    @utils.add_log
    def split_bam(self):
        barcode_tag_dict = {}
        tags = self.tag_barcode_dict.keys()
        for tag in tags:
            for barcode in self.tag_barcode_dict[tag]:
                barcode_tag_dict[barcode] = tag

        with pysam.AlignmentFile(self.args.bam_file, 'r') as bam_in:
            bam_handle_dict = {}
            for tag in tags:
                bam_file = f'{self.bam_outdir}/{tag}.bam'
                bam_handle_dict[tag] = pysam.AlignmentFile(bam_file, "wb", header=bam_in.header)
            
            for seg in bam_in:
                if not seg.has_tag('XT'):
                    continue
                barcode = seg.get_tag('CB')
                if barcode in barcode_tag_dict:
                    tag = barcode_tag_dict[barcode]
                    bam_handle_dict[tag].write(seg)

        for h in bam_handle_dict.values():
            h.close()
            
    @utils.add_log
    def split_vdj(self):

        df_vdj = pd.read_csv(self.cell_confident_vdj, sep='\t')
        for tag in self.tag_barcode_dict:
            tag_barcodes = list(self.tag_barcode_dict[tag])
            temp = df_vdj[df_vdj.barcode.isin(tag_barcodes)]
            if not temp.empty:
                clonotypes, seqtype = get_clonotypes_table(temp)
                temp.to_csv(f'{self.vdj_outdir}/{tag}_{seqtype}_cell_confident.tsv', sep='\t', index=False)
                clonotypes.to_csv(f'{self.vdj_outdir}/{tag}_{seqtype}_clonotypes.tsv', sep='\t', index=False)
            else:
                continue
    
    @utils.add_log
    def split_fl_vdj(self):
        df_anno = pd.read_csv(self.anno_file, keep_default_na=False)
        if df_anno.chain[0].startswith('IG'):
            seqtype = 'BCR'
        else:
            seqtype = 'TCR'

        for tag in self.tag_barcode_dict:
            tag_barcodes = set(self.tag_barcode_dict[tag])
            df_temp = df_anno[df_anno.barcode.isin(tag_barcodes)]
            if not df_temp.empty:
                df_temp.to_csv(f'{self.fl_vdj_outdir}/{tag}_{seqtype}_contig_annotations.csv', sep=',', index=False)
                fasta_temp = open(f'{self.fl_vdj_outdir}/{tag}_{seqtype}_contig.fasta' ,'w')
                with pysam.FastxFile(self.fasta_file) as raw_fasta:
                    for entry in raw_fasta:
                        name = entry.name
                        attrs = name.split(':')
                        cb = attrs[0]
                        if cb in tag_barcodes:
                            seq = entry.sequence
                            fasta_temp.write(f'>{name}\n{seq}\n')
                fasta_temp.close()

                if 'clonotype_id' in df_temp.columns:
                    df_temp = df_temp[df_temp.clonotype_id != ''] # del clonotype for trust4
                split_clonotypes = f'{self.fl_vdj_outdir}/{tag}_{seqtype}_clonotypes.csv'

                gen_clonotypes_table(df_temp, split_clonotypes, seqtype)
                metrics_dict = gen_vj_annotation_metrics(df_temp, seqtype)
                utils.dump_dict_to_json(metrics_dict, f"{self.fl_vdj_outdir}/{tag}_metrics.json")


    @utils.add_log
    def run(self):
        if self.args.split_matrix:
            self.split_matrix()
        if self.args.split_bam:
            self.split_bam()
        if self.args.split_fastq:
            self.write_r2_fastq_files()
            self.write_r1_fastq_files()
        if self.args.split_vdj:
            self.split_vdj()
        if self.args.split_fl_vdj:
            self.split_fl_vdj()


def split_tag(args):
    with Split_tag(args) as runner:
        runner.run()


def get_opts_split_tag(parser, sub_program):
    parser.add_argument(
        "--split_fastq",
        help="If used, will split scRNA-Seq fastq file according to tag assignment.",
        action='store_true',
    )
    parser.add_argument(
        "--split_matrix",
        help="If used, will split scRNA-Seq matrix file according to tag assignment.",
        action='store_true',
    )
    parser.add_argument(
        "--split_bam",
        help="If used, will split scRNA-Seq featureCounts bam file according to tag assignment. Use together with --bam_file",
        action='store_true',
    )
    parser.add_argument("--bam_file", help='scRNA-Seq bam file to split.')
    parser.add_argument(
        "--split_vdj",
        help="If used, will split scRNA-Seq vdj count file according to tag assignment.",
        action='store_true',
    )
    parser.add_argument(
        "--split_fl_vdj",
        help='If used, will split scRNA-Seq full-length vdj annotation, fasta, clonotypes file according to tag assignment.',
        action='store_true',
    )
    parser.add_argument("--vdj_dir", help="Match celescope vdj directory. Required when --split_vdj or --split_fl_vdj is specified.")
    if sub_program:
        parser.add_argument("--umi_tag_file", help="UMI tag file.", required=True)
        parser.add_argument("--match_dir", help=HELP_DICT['match_dir'])
        parser.add_argument("--matrix_dir", help="Match celescope scRNA-Seq matrix directory.")
        parser.add_argument("--R1_read", help='R1 read path.')
        s_common(parser)



   
import math

import matplotlib.pyplot as plt
import matplotlib
import numpy as np

import celescope.tools.utils as utils

matplotlib.use('Agg')


class Otsu():
    """
    remove all zero in array
    Return 1 if len(array) < otsu_min_len
    """
    def __init__(self, array, log_base=10, otsu_min_len = 50, otsu_plot_path=None, **kwargs):
        
        self.len_bool = True
        array = [x for x in array if x > 0 ]
        if len(array) < otsu_min_len:
            self.len_bool = False

        self.log_base = int(log_base)
        # base change rule
        self.array = np.log(array) / np.log(self.log_base)

        self.kwargs = kwargs
        self.threshold = 1
        self.counts = None
        self.bins = None
        self.otsu_plot_path = otsu_plot_path


    def _threshold_otsu(self):
        """Return threshold value based on Otsu's method.
        hist : array, or 2-tuple of arrays, optional
            Histogram from which to determine the threshold, and optionally a
            corresponding array of bin center intensities.
            An alternative use of this function is to pass it only hist.
        Returns
        -------
        threshold : float
            Upper threshold value. All pixels with an intensity higher than
            this value are assumed to be foreground.
        References
        ----------
        .. [1] Wikipedia, https://en.wikipedia.org/wiki/Otsu's_Method
        """
        counts = self.counts
        bin_centers = self.bins[:-1]

        # class probabilities for all possible thresholds
        weight1 = np.cumsum(counts)
        weight2 = np.cumsum(counts[::-1])[::-1]
        # class means for all possible thresholds
        mean1 = np.cumsum(counts * bin_centers) / weight1
        mean2 = (np.cumsum((counts * bin_centers)[::-1]) / weight2[::-1])[::-1]

        # Clip ends to align class 1 and class 2 variables:
        # The last value of ``weight1``/``mean1`` should pair with zero values in
        # ``weight2``/``mean2``, which do not exist.
        variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:]) ** 2
        if len(variance12) == 0:
            return 0
        idx = np.nanargmax(variance12)
        self.threshold = bin_centers[idx]

    def _array2hist(self, binWidth=0.2):
        self.counts, self.bins = np.histogram(self.array, bins=np.arange(0, max(self.array)+binWidth, binWidth))

    def _make_plot(self):
        plt.hist(x=self.bins[:-1], bins=self.bins, weights=self.counts)
        plt.axvline(self.threshold, color='r')
        plt.xlabel(f'log{self.log_base} observed counts')
        plt.ylabel('Frequency')
        plt.savefig(self.otsu_plot_path)
        plt.close()

    def run(self):
        """
        return threshold
        """
        if not self.len_bool:
            return 1
        self._array2hist()
        self._threshold_otsu()
        if self.otsu_plot_path:
            self._make_plot()
        return_threshold = math.ceil(self.log_base ** self.threshold)

        return return_threshold


class Auto():
    """
    threshold = top {percentile}% cell count / coef
    count is usually UMI count.
    >>> array = [50] * 100 + [30] * 100 + [10] * 100 + [4] * 100
    >>> Auto(array, coef=10).run()
    5
    >>> Auto(array, percentile=70, coef=3).run()
    10
    >>> Auto(array, percentile=50, coef=10, expected_cell_num=100).run()
    5
    >>> Auto([1, 2, 20, 30, 40], expected_cell_num=4, percentile=50, coef=10).run()
    2
    """
    def __init__(self, array, percentile=99, coef=3, expected_cell_num=None, **kwargs):
        self.array = [x for x in array if x > 0 ]
        self.percentile = percentile
        self.coef = int(coef)
        self.expected_cell_num = expected_cell_num
        self.kwargs = kwargs
    
    def run(self):
        array = self.array
        if not array:
            return 1

        if not self.expected_cell_num:
            expected_cell_num = len(array)
        else:
            expected_cell_num = self.expected_cell_num
            if expected_cell_num > len(array):
                print('Warning: expected_cell_num > len(array)')
                expected_cell_num = len(array)
                      
        sorted_counts = sorted(array, reverse=True)
        count_cell_percentile = np.percentile(sorted_counts[:expected_cell_num], self.percentile)
        threshold = int(count_cell_percentile / self.coef)

        return threshold


class Threshold():
    """
    Args:
        array: array-like
        threshold_method: ['otsu', 'auto', 'hard', 'none']
        otsu_plot_path: str
        hard_threshold: int
    """
    def __init__(self, array, threshold_method='auto', otsu_plot_path=None, hard_threshold=None, **kwargs):
        self.array = [x for x in array if x > 0 ]
        self.threshold_method = threshold_method
        self.otsu_plot_path = otsu_plot_path
        self.hard_threshold = hard_threshold

        self.kwargs = kwargs
    
    @utils.add_log
    def run(self):
        """
        return threshold
        """
        if not self.array:
            return 1
        if self.threshold_method == 'otsu':
            otsu = Otsu(self.array, otsu_plot_path=self.otsu_plot_path, **self.kwargs)
            threshold = otsu.run()
        elif self.threshold_method == 'auto':
            auto = Auto(self.array, **self.kwargs)
            threshold = auto.run()
        elif self.threshold_method == 'hard':
            if self.hard_threshold:
                threshold = int(self.hard_threshold)
            else:
                raise Exception('hard_threshold must be set')
        elif self.threshold_method == 'none':
            threshold = 1
        else:
            raise ValueError(f'Unknown threshold method: {self.threshold_method}')

        return threshold


import subprocess
import json

from celescope.tools.step import Step, s_common
from celescope.tools import utils

POLY_A = '-a polyA=A{18} '

LOG_METRICS_TITLE = (
    'Total reads processed',
    'Reads with adapters',
    'Reads that were too short',
    'Reads written (passing filters)',
    'Total basepairs processed',
    'Quality-trimmed',
    'Total written (filtered)',
)


def get_cutadapt_cmd(args, input_file, output_file):
    cmd = (
        'cutadapt '
        f'{POLY_A} '
        f'-j {args.thread} '
        f'-m {args.minimum_length} '
        f'--nextseq-trim={args.nextseq_trim} '
        f'--overlap {args.overlap} '
        f'{args.cutadapt_param} '
        f'--json {args.outdir}/{args.sample}_cutadapt.json '
        f'-o {output_file} '
        f'{input_file} '
        )
    return cmd

class Cutadapt(Step):
    """
    ## Features
    - Trim poly A tails and user-provided adapters in R2 reads with cutadapt.

    ## Output
    - `cutadapt.log` Cutadapt output log file.
    - `{sample}_clean_2.fq.gz` R2 reads file without adapters.
    """

    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

        # out files
        self.out_fq2 = f'{self.outdir}/{self.sample}_clean_2.fq'
        self.json_log = f'{self.outdir}/{self.sample}_cutadapt.json'


    def add_cutadapt_metrics(self):
        with open(self.json_log) as f:
            log_dict = json.load(f)

        total_reads = log_dict['read_counts']['input']
        reads_with_adapters = log_dict['read_counts']['read1_with_adapter']
        reads_too_short = log_dict['read_counts']['filtered']['too_short']
        total_base_pairs = log_dict['basepair_counts']['input']
        quality_trimmed = log_dict['basepair_counts']['quality_trimmed']

        self.add_metric(
            name='Reads with Adapters',
            value=reads_with_adapters,
            total=total_reads,
            help_info='reads with poly A tails and user-provided adapters(if any) are trimmed'
        )
        self.add_metric(
            name='Reads too Short',
            value=reads_too_short,
            total=total_reads,
            help_info=f'reads with read length less than {self.args.minimum_length} bp after trimming'
        )
        self.add_metric(
            name='Base Pairs Quality-Trimmed',
            value=quality_trimmed,
            total=total_base_pairs,
            help_info='bases pairs removed from the end of the read whose quality is smaller than the given threshold'
        )

    @utils.add_log
    def run(self):
        input_file = self.args.fq
        output_file = self.out_fq2
        cmd = get_cutadapt_cmd(self.args, input_file, output_file)
        self.run.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)
        self.add_cutadapt_metrics()


@utils.add_log
def cutadapt(args):
    with Cutadapt(args, display_title="Trimming") as runner:
        runner.run()


def get_opts_cutadapt(parser, sub_program):
    parser.add_argument(
        '--minimum_length',
        help='Discard processed reads that are shorter than LENGTH.',
        default=20
    )
    parser.add_argument(
        '--nextseq_trim',
        help="""Quality trimming of reads using two-color chemistry (NextSeq). 
Some Illumina instruments use a two-color chemistry to encode the four bases. 
This includes the NextSeq and the NovaSeq. 
In those instruments, a ‘dark cycle’ (with no detected color) encodes a G. 
However, dark cycles also occur when sequencing “falls off” the end of the fragment.
The read then contains a run of high-quality, but incorrect “G” calls at its 3’ end.""",
        default=20,
    )
    parser.add_argument(
        '--overlap',
        help="""Since Cutadapt allows partial matches between the read and the adapter sequence,
short matches can occur by chance, leading to erroneously trimmed bases. 
For example, roughly 0.25 of all reads end with a base that is identical to the first base of the adapter. 
To reduce the number of falsely trimmed bases, the alignment algorithm requires that 
at least {overlap} bases match between adapter and read. """,
        default=10
    )
    parser.add_argument('--cutadapt_param', help='Other cutadapt parameters. For example, --cutadapt_param "-a p5=AGATCGGAAGAGCACACGTCTGAACTCCAGTCA" ', default="")
    if sub_program:
        parser.add_argument('--fq', help='Required. R2 reads from step Barcode.', required=True)
        parser = s_common(parser)
    return parser



import pandas as pd
import numpy as np
import numbers
import pysam
from collections import defaultdict
from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.vdj.__init__ import CHAINS


INDEX_DICT = {
    'CTCCAT': 'TS-B1',
    'ATCCTC': 'TS-B2',
    'ACGTCT': 'TS-B3',
    'TGCGAA': 'TS-B4',
    'TTCTCG': 'TS-B5',
    'CTGCTA': 'TS-B6',
    'CATGAT': 'TS-B7',
    'TCAACT': 'TS-B8',
    'AGTCCT': 'TS-B9',
    'GTTGAG': 'TS-B10',
    'TAGCTG': 'TS-B11',
    'TCGCCA': 'TS-B12',
    'GAACTC': 'TS-B13',
    'TATGGT': 'TS-B14',
    'CGCAAC': 'TS-B15',
    'TGGCAG': 'TS-B16',
    'ATGCAT': 'TS-B17',
    'GACTAT': 'TS-B18',
    'GTGATT': 'TS-B19',
    'CTCTTG': 'TS-B20',
    'AAGCGT': 'TS-B21',
    'CCAACA': 'TS-B22',
    'GTTGGT': 'TS-B23',
    'TCTAGT': 'TS-B24',
    'TATGTG': 'TS-B25',
    'TGTGGC': 'TS-B26',
    'GACCTG': 'TS-B27',
    'TTCCGT': 'TS-B28',
    'AAGGCA': 'TS-B29',
    'TAGGAT': 'TS-B30',
    'AACTCC': 'TS-B31',
    'TCGATG': 'TS-B32',
    'CTGCGT': 'TS-B33',
    'GTCGGA': 'TS-B34',
    'TCACAT': 'TS-B35',
    'ATAGGT': 'TS-B36',
    'CGTAAT': 'TS-B37',
    'GCATGT': 'TS-B38',
    'AACTGA': 'TS-B39',
    'GCACAA': 'TS-B40',
    'GCCATC': 'TS-B41',
    'CAACCG': 'TS-B42',
    'GTCTGG': 'TS-B43',
    'TCCATT': 'TS-B44',
    'CAGACC': 'TS-B45',
    'ACGGAG': 'TS-B46',
    'ACATCA': 'TS-B47',
    'TATCCG': 'TS-B48',
    'GGAGAG': 'TS-B49',
    'CCAATG': 'TS-B50',
    'TTCTGA': 'TS-B51',
    'GTGACG': 'TS-B52',
    'ATGGTG': 'TS-B53',
    'ACTTGT': 'TS-B54',
    'ATAGAC': 'TS-B55',
    'CCTATA': 'TS-B56',
    'TTAAGG': 'TS-B57',
    'GATCAC': 'TS-B58',
    'TAGCCT': 'TS-B59',
    'AGCGCT': 'TS-B60',
    'AGACGC': 'TS-B61',
    'CTAAGA': 'TS-B62',
    'TATCGA': 'TS-B63',
    'CGCACA': 'TS-B64',
    'CAAGTT': 'TS-B65',
    'GAACCA': 'TS-B66',
    'TACACA': 'TS-B67',
    'CATTGG': 'TS-B68',
    'TCATGC': 'TS-B69',
    'AGGTTA': 'TS-B70',
    'TCGAAT': 'TS-B71',
    'TCTTGG': 'TS-B72',
    'CTCTAC': 'TS-B73',
    'GAGGTC': 'TS-B74',
    'ACAACG': 'TS-B75',
    'CAGATA': 'TS-B76',
    'CAGGTA': 'TS-B77',
    'TCTTAC': 'TS-B78',
    'CCTGTG': 'TS-B79',
    'TCGAGC': 'TS-B80',
    'CTGAAT': 'TS-B81',
    'ATTGGC': 'TS-B82',
    'CATCTT': 'TS-B83',
    'TCTCTA': 'TS-B84',
    'GCGTCA': 'TS-B85',
    'GTTCAT': 'TS-B86',
    'AATCAG': 'TS-B87',
    'CGGTGT': 'TS-B88',
    'TCCGTC': 'TS-B89',
    'CTCACC': 'TS-B90',
    'TTGACT': 'TS-B91',
    'GCCGTA': 'TS-B92',
    'CGACTC': 'TS-B93',
    'ATCCAA': 'TS-B94',
    'TGCCAT': 'TS-B95',
    'ACGATA': 'TS-B96',
}


def format_value(value, total):
    if not isinstance(value, numbers.Number):
        return value
    display = str(format(value, ','))
    if total:
        fraction = round(value / total * 100, 2)
        display += f'({fraction}%)'
    return display


def correct_cdr3_nt(umi_dict, percent=0.1):
    """
    Correct umi_dict in place.
    Args:
        umi_dict: {cdr3_nt: umi_count}
        percent: if hamming_distance(low_umi_cdr3, high_umi_cdr3) == 1 and
            low_count / high_count < percent, merge low to high.
    Returns:
        correct_dict: dict {low_umi_cdr3: high_umi_cdr3}
    """
    correct_dict = dict()
    
    umi_arr = sorted(
        umi_dict.items(), key=lambda kv: (kv[1], kv[0]), reverse=True)
    
    while True:
        # break when only highest in umi_arr
        if len(umi_arr) <= 1:
            break
        umi_low = umi_arr.pop()
        low_seq = umi_low[0]
        low_count = umi_low[1]

        for umi_kv in umi_arr:
            high_seq = umi_kv[0]
            high_count = umi_kv[1]
            if len(low_seq) != len(high_seq):
                break
            if float(low_count / high_count) > percent:
                break
            if utils.hamming_distance(low_seq, high_seq) == 1:
                correct_dict[low_seq] = high_seq
                n_low = umi_dict[low_seq]
                # merge
                umi_dict[high_seq] += n_low
                del (umi_dict[low_seq])
                break
            
    return correct_dict


def simpson_di(data):

    """ Given a hash { 'species': count } , returns the Simpson Diversity Index
    
    >>> simpson_di({'a': 10, 'b': 20, 'c': 30,})
    0.3888888888888889
    """

    def p(n, N):
        """ Relative abundance """
        if n == 0:
            return 0
        else:
            return float(n)/N

    N = sum(data.values())
    
    return sum(p(n, N)**2 for n in data.values() if n != 0)


def inverse_simpson_di(data):
    """ Given a hash { 'species': count } , returns the inverse Simpson Diversity Index
    
    >>> inverse_simpson_di({'a': 10, 'b': 20, 'c': 30,})
    2.571428571428571
    """
    return float(1)/simpson_di(data)


class Count_vdj(Step):
    """
    ## Features  
    - Summarize clonetypes infomation.
    ## Output
    - `{sample}_mapping_metrics.tsv` Mapping metrics of each index.
    - `{sample}_corrected_productive.tsv` Productive chains after cdr3_nt correcting.
    - `{sample}_filtered_annotations.csv` Annotation for each CDR3.
    - `{sample}_clonetypes.csv` The count and percentage of each CDR3_aa of umis.
    - `{sample}_clonetypes_by_nt.csv` The count and percentage of each CDR3_nt of umis.
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.chains = CHAINS[args.type]

        # IN
        self.airr_file = args.airr_file
        self.productive_file = args.productive_file
        self.fq = args.fq

        # OUT
        self.mapping_result_file = f"{self.out_prefix}_mapping_metrics.tsv"
        self.corrected_productive_file = f"{self.out_prefix}_corrected_productive.tsv"
        self.annotation = f"{self.out_prefix}_filtered_annotations.csv"
        self.clonetypes_aa = f"{self.out_prefix}_clonetypes.csv"
        self.clonetypes_nt = f"{self.out_prefix}_clonetypes_nt.csv"
    
    @utils.add_log
    def gen_mapping_metrics(self):
        """mapping metrics for each index.
        """
        df = pd.read_csv(self.airr_file, sep='\t')
        df.fillna("", inplace=True)
        df["index"] = df["sequence_id"].apply(lambda x: x.split(':')[0])
        
        df_metrics = pd.DataFrame(columns=[
            "Index", "UMI_Counts", "UMIs_Mapped_To_Any_VDJ_Gene", "UMIs_with_CDR3",
            "UMIs_with_Correct_CDR3", "UMIs_Mapped_Confidently_To_VJ_Gene",
        ])
        for chain in self.chains:
            df_metrics[f"UMIs_Mapped_To_{chain}"] = ''
            
        for index, data in df.groupby("index"):
            umi_count = data.shape[0]
            
            data = data[(data["v_call"]!="") | ((data["d_call"]!="")) | ((data["j_call"]!=""))]
            UMIs_Mapped_To_Any_VDJ_Gene = data.shape[0]

            df_cdr3 = data[(data["cdr3_aa"]!="") & (data["junction_aa"]!="")]
            UMIs_with_CDR3 = df_cdr3.shape[0]
            
            df_correct_cdr3 = df_cdr3[~(df_cdr3["cdr3_aa"].str.contains(r"\*")) & ~(df_cdr3["cdr3_aa"].str.contains("X"))]
            UMIs_with_Correct_CDR3 = df_correct_cdr3.shape[0]
            
            df_confident = df_correct_cdr3[df_correct_cdr3["productive"]=="T"]
            UMIs_Mapped_Confidently_To_VJ_Gene = df_confident.shape[0]
            
            UMIs_Mapped_To_chains = []
            for chain in self.chains:
                df_chain = df_confident[df_confident.locus == chain]
                UMIs_Mapped_To_chains.append(df_chain.shape[0])
            
            metrics_list = [
                index, umi_count, UMIs_Mapped_To_Any_VDJ_Gene, UMIs_with_CDR3,
                UMIs_with_Correct_CDR3, UMIs_Mapped_Confidently_To_VJ_Gene,
            ]
            metrics_list += UMIs_Mapped_To_chains
            metrics_list = [format_value(value, umi_count) for value in metrics_list]
            df_metrics.loc[len(df_metrics.index)] = metrics_list

        readcount_dict = defaultdict(int)
        with pysam.FastxFile(self.fq) as f:
            for read in f:
                index = read.name.split(':')[0]
                readcount_dict[index] += 1
        
        df_metrics.insert(1, "read_count", df_metrics["Index"].apply(lambda x: readcount_dict[x]))
        df_metrics.sort_values("read_count", ascending=False, inplace=True)
        df_metrics["read_count"] = df_metrics["read_count"].apply(lambda x: format(x, ','))
        df_metrics["Index"] = df_metrics["Index"].apply(lambda x: INDEX_DICT[x])
        df_metrics.to_csv(self.mapping_result_file, sep="\t", index=False)
        
    @utils.add_log
    def correct_cdr3_nt(self):
        """
        Correct cdr3_nt sequence.
        """
        self.productive_file = pd.read_csv(self.productive_file, sep='\t')
        self.productive_file = self.productive_file[self.productive_file["chain"].isin(self.chains)]
        self.productive_file["barcode"] = self.productive_file["barcode"].apply(lambda x: INDEX_DICT[x])
        self.productive_file["umi"] = self.productive_file["sequence_id"].apply(lambda x: x.split(':')[1])

        for chain in self.chains:
            df_tmp = self.productive_file[self.productive_file["chain"] == chain]
            groupby_elements = ["chain","nSeqCDR3"]
            clonetypes = df_tmp.groupby(groupby_elements, as_index=False).agg({"umi": "count"})
            clonetypes = clonetypes.sort_values("umi", ascending=False)
            umi_dict = dict(zip(list(clonetypes.nSeqCDR3),list(clonetypes.umi)))
            correct_dict = correct_cdr3_nt(umi_dict)
    
            for low_nt, high_nt in correct_dict.items():
                high_aa = df_tmp.loc[df_tmp.nSeqCDR3 == high_nt].aaSeqCDR3.iloc[0]
                self.productive_file.loc[self.productive_file.nSeqCDR3 == low_nt, 'aaSeqCDR3'] = high_aa
                self.productive_file.loc[self.productive_file.nSeqCDR3 == low_nt, 'nSeqCDR3'] = high_nt
                
        self.productive_file.to_csv(self.corrected_productive_file, sep="\t", index=False)
    
    @utils.add_log
    def gen_clonotype_file(self):
        """Output clonotypes.csv file.  
        
        ClonotypeID	aaSeqCDR3	Frequency	Proportion	Index	diversity
        1	TRB:CASGDRLGGSQNTLYF	105	2.34%	ACTTGT	463.14
        2	TRA:CAASANSGTYQRF	36	0.80%	ACTTGT	463.14
        3	TRA:CALSRNTGYQNFYF	30	0.67%	ACTTGT	463.14
        4	TRA:CAASADYGNEKITF	30	0.67%	ACTTGT	463.14
        5	TRA:CAASVDNYAQGLTF	27	0.60%	ACTTGT	463.14
        """
        index_set = set(self.productive_file.barcode)
        cdr3_types = ["nSeqCDR3", "aaSeqCDR3"]
        final_df_aa = pd.DataFrame()
        final_df_nt = pd.DataFrame()
        
        def format_files(index, cdr3_type):
            """format clonotypes.csv and clonotypes_nt.csv from productive file."""
            nonlocal final_df_aa, final_df_nt
            
            groupby_elements = ["chain", cdr3_type]
            df_clonetypes = self.productive_file[self.productive_file["barcode"]==index]
            df_clonetypes = df_clonetypes.groupby(groupby_elements, as_index=False).agg({"umi": "count"})
            df_clonetypes[cdr3_type] = df_clonetypes.loc[:, ["chain", cdr3_type]].apply(':'.join, axis=1)
            df_clonetypes = df_clonetypes.sort_values("umi", ascending=False).reset_index()
            df_clonetypes["ClonotypeID"] = pd.Series(df_clonetypes.index) + 1
            df_clonetypes = df_clonetypes.rename(columns={"umi": "Frequency"})
    
            sum_frequency = sum(df_clonetypes["Frequency"])
            df_clonetypes["Proportion"] = df_clonetypes["Frequency"].apply(lambda x : x / sum_frequency)
            df_clonetypes["Proportion"] = df_clonetypes["Proportion"].apply(lambda x: str(round(x*100, 2)) + '%' )
            df_clonetypes = df_clonetypes[["ClonotypeID", cdr3_type, "Frequency", "Proportion"]]
            df_clonetypes["Index"] = index

            # calculate diversity
            data = dict(zip(df_clonetypes[cdr3_type], df_clonetypes["Frequency"]))
            clonotype_diversity = round(inverse_simpson_di(data), 2)
            df_clonetypes["Diversity"] = clonotype_diversity
            # df_clonetypes.loc[df_clonetypes.Index == index, "Diversity"] = clonotype_diversity
            if cdr3_type == "aaSeqCDR3":
                final_df_aa = pd.concat([final_df_aa, df_clonetypes])
            else:
                final_df_nt = pd.concat([final_df_nt, df_clonetypes])

        for index in index_set:
            for cdr3_type in cdr3_types:
                format_files(index, cdr3_type)

        final_df_aa.to_csv(self.clonetypes_aa, sep=',', index=False)
        final_df_nt.to_csv(self.clonetypes_nt, sep=',', index=False)
        
        df_table = final_df_aa.groupby('Index').head(50)
        df_table = df_table[["Index", "ClonotypeID", "aaSeqCDR3", "Frequency", "Proportion", "Diversity"]]
        
        mean_diversity = round(np.median(df_table.Diversity), 2)
        self.add_metric(
            name="Median Diversity",
            value=mean_diversity,
            help_info="Median inverse Simpson index of all wells"
        )
        
        table_dict = self.get_table_dict(
            title="Clonotypes by well index",
            table_id='bulk_vdj',
            df_table=df_table
        )
        self.add_data(table_dict=table_dict)


    @utils.add_log
    def gen_annotation_file(self):
        """
        Output annotation file
        """
        in_file = self.productive_file.copy()
        
        groupby_elements = ["barcode", "chain", "aaSeqCDR3"]
        in_file['umis'] = in_file.groupby(groupby_elements)['umi'].transform('count')

        out_file = pd.DataFrame({
                        "barcode":in_file.barcode,
                        "is_cell":True,
                        "sequence_id":in_file.sequence_id,
                        "high_confidence":True,
                        "chain":in_file.chain,
                        "v_gene":in_file.bestVGene,
                        "d_gene":in_file.bestDGene,
                        "j_gene":in_file.bestJGene,
                        "c_gene":'None',
                        "full_length":True,
                        "productive":True,
                        "cdr3":in_file.aaSeqCDR3,
                        "cdr3_nt":in_file.nSeqCDR3,
                        "reads":in_file.umis,
                        "umis":in_file.umis,
                        })
        out_file.fillna("None", inplace=True)
        out_file.to_csv(self.annotation, sep=',', index=False)

    def run(self):
        self.gen_mapping_metrics()
        self.correct_cdr3_nt()
        self.gen_clonotype_file()
        self.gen_annotation_file()


def count_vdj(args):
    with Count_vdj(args, display_title="Clonotypes") as runner:
        runner.run()


def get_opts_count_vdj(parser, sub_program):
    parser.add_argument(
        "--type", help="Required. `TCR` or `BCR`. ", required=True)
    if sub_program:
        parser.add_argument("--productive_file", help="Required. Productive file from mapping_vdj step.", required=True)
        parser.add_argument("--airr_file", help="Required. Airr file from mapping_vdj step.", required=True)
        parser.add_argument("--fq", help="Required. R2 reads from step Barcode.", required=True)
        parser = s_common(parser)


'''
vdj mapping
'''
import pandas as pd
import subprocess
import pysam
from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.vdj.__init__ import CHAINS
from xopen import xopen


SPLIT_N_CHUNKS = 4


class Mapping_vdj(Step):
    """
    ## Features
    - Align R2 reads to IGMT(http://www.imgt.org/) database sequences with blast.

    ## Output
    - `{sample}_airr.tsv` The alignment result of each UMI.
    A tab-delimited file compliant with the AIRR Rearrangement schema(https://docs.airr-community.org/en/stable/datarep/rearrangements.html)

    - `{sample}_UMI_count_unfiltered.tsv` UMI reading for each (barcode, chain, VJ_pair) combination.

    - `{sample}_UMI_count_filtered.tsv` For each (barcode, chain) combination, only the record with the 
    most VJ_pair UMI reads is kept.

    """
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.ref_path = args.ref_path
        self.seqtype = args.type
        self.species = args.species
        self.split_fasta = args.split_fasta
        self.chains = CHAINS[self.seqtype]

        # out
        self.UMI_count_unfiltered_file = f'{self.out_prefix}_UMI_count_unfiltered.tsv'
        self.UMI_count_filtered_file = f'{self.out_prefix}_UMI_count_filtered.tsv'
        
        if self.split_fasta:
            utils.check_mkdir(f'{self.outdir}/tmp/')
            self.UMI_number = int(self.get_slot_key(slot='metrics', step_name='consensus', key='UMI Counts'))
            self.tmp_fasta = [f'{self.outdir}/tmp/{self.sample}_{i}.fasta' for i in range(SPLIT_N_CHUNKS)]
            self.airr_out = [f'{self.out_prefix}_airr_{i}.tsv' for i in range(SPLIT_N_CHUNKS)]
            
        else:
            self.airr_out = [f'{self.out_prefix}_airr.tsv']
    
    def run(self):
        # run igblstn
        if not self.split_fasta:
            self.igblast(fasta=self.args.fasta, airr_out=self.airr_out[0])
        else:
            self.split_fasta_file()
            for i in range(len(self.tmp_fasta)):
                self.igblast(fasta=self.tmp_fasta[i], airr_out=self.airr_out[i])
                
        self.mapping_summary()

    @utils.add_log
    def split_fasta_file(self):
        fh_tmp_fasta = [xopen(i, 'w') for i in self.tmp_fasta]
        umi_count = 0
        chunk = self.UMI_number // SPLIT_N_CHUNKS
        with pysam.FastxFile(self.args.fasta) as f:
            for read in f:
                umi_count += 1
                if umi_count <= chunk:
                    fh_tmp_fasta[0].write(utils.fasta_line(read.name, read.sequence))
                elif chunk < umi_count <= 2*chunk:
                    fh_tmp_fasta[1].write(utils.fasta_line(read.name, read.sequence))
                elif 2*chunk < umi_count <= 3*chunk:
                    fh_tmp_fasta[2].write(utils.fasta_line(read.name, read.sequence))
                else:
                    fh_tmp_fasta[3].write(utils.fasta_line(read.name, read.sequence))
        
        for i in fh_tmp_fasta:
            i.close()
        
    @utils.add_log
    def igblast(self, fasta, airr_out):
        if self.seqtype == "TCR":
            chain, ig_seqtype ="TR", "TCR"
        else:
            chain, ig_seqtype ="IG", "Ig"
            
        cmd = (
            f"igblastn -query {fasta} "
            f"-organism {self.species} "
            f"-ig_seqtype {ig_seqtype} "
            f"-auxiliary_data optional_file/{self.species}_gl.aux "
            f"-num_threads {self.args.thread} "
            f"-germline_db_V {self.ref_path}/{chain}V.fa "
            f"-germline_db_D {self.ref_path}/{chain}D.fa "
            f"-germline_db_J {self.ref_path}/{chain}J.fa "
            "-domain_system imgt -show_translation -outfmt 19 " # outfmt19 is an AIRR tab-delimited file, IgBLAST v1.9.0 or higher required.
            f"-out {airr_out} "
        )

        self.igblast.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def mapping_summary(self):
        self.add_metric(
            name="Species",
            value=self.species,
            help_info="Human or Mouse"
        )
        
        df_total_confident = pd.DataFrame()
        total_reads, map_to_any_vdj_gene_num, cdr3_num, correct_cdr3_num, confident_num = 0, 0, 0, 0, 0
        
        for df in self.airr_out:
            df = pd.read_csv(df, sep='\t')
            df.fillna("", inplace=True)
            total_reads += df.shape[0]

            # mapping to any vdj genes
            df = df[(df["v_call"]!="") | ((df["d_call"]!="")) | ((df["j_call"]!=""))]
            map_to_any_vdj_gene_num += df.shape[0]
            # UMIs with CDR3
            df = df[df["cdr3_aa"]!=""]
            cdr3_num += df.shape[0]
            # UMIs with Correct CDR3
            df = df[~(df["cdr3_aa"].str.contains(r"\*")) & ~(df["cdr3_aa"].str.contains("X"))]
            correct_cdr3_num += df.shape[0]
            # UMIs Mapped Confidently To VJ Gene
            df = df[df["productive"]=="T"]
            confident_num += df.shape[0]
            
            df_total_confident = pd.concat([df_total_confident, df])
            
        self.add_metric(
            name="UMIs Mapped to Any VDJ Gene",
            value=map_to_any_vdj_gene_num,
            total=total_reads,
            help_info="UMI Mapped to any germline VDJ gene segments"
        )
        self.add_metric(
            name="UMIs with CDR3",
            value=cdr3_num,
            total=total_reads,
            help_info="UMIs with CDR3 sequence"
        )
        self.add_metric(
            name="UMIs with Correct CDR3",
            value=correct_cdr3_num,
            total=total_reads,
            help_info="UMIs with CDR3 might have stop codon and these UMIs(or Reads) are classified as incorrect"
        )
        self.add_metric(
            name="UMIs Mapped Confidently to VJ Gene",
            value=confident_num,
            total=total_reads,
            help_info="UMIs with productive rearrangement mapped to VJ gene pairs and with correct CDR3"
        )

        # UMIs Mapped Confidently to each chain
        for chain in self.chains:
            df_chain = df_total_confident[df_total_confident.locus == chain]
            self.add_metric(
                name=f"UMIs Mapped to {chain}",
                value=df_chain.shape[0],
                total=total_reads,
                help_info=f"UMIs mapped confidently to {chain}"
            )

        # output file
        df_VJ = df_total_confident[["sequence_id","locus", "v_call", "d_call", "j_call", "junction", "junction_aa"]]
        df_VJ.rename(columns={
            "sequence_id": "readID",
            "locus": "chain",
            "v_call": "bestVGene",
            "d_call": "bestDGene",
            "j_call": "bestJGene",
            "junction": "nSeqCDR3",
            "junction_aa": "aaSeqCDR3"
        }, inplace=True)
        df_VJ["barcode"] = df_VJ["readID"].apply(lambda x: x.split(':')[0])
        df_VJ["UMI"] = df_VJ["readID"].apply(lambda x: x.split(':')[1])

        for i in ["bestVGene", "bestDGene", "bestJGene"]:
            df_VJ[i] = df_VJ[i].apply(lambda x: x.split("*")[0])

        # filter1: keep top 1 in each combinations
        groupby_elements = [
            'barcode',
            'chain',
            'bestVGene',
            'bestJGene',
            'aaSeqCDR3',
            'nSeqCDR3',
        ]

        # unique UMI
        df_UMI = df_VJ.drop_duplicates(subset=["barcode", "UMI"], keep="first")
        df_UMI_count = df_UMI.groupby(
            groupby_elements, as_index=False).agg({"UMI": "count"})
        df_UMI_count = df_UMI_count.sort_values("UMI", ascending=False)
        # out unfiltered
        df_UMI_count.to_csv(self.UMI_count_unfiltered_file, sep="\t", index=False)

        df_UMI_count_filter = df_UMI_count.groupby(
            ["barcode", "chain"], as_index=False).head(1)
        # out filtered
        df_UMI_count_filter.to_csv(
            self.UMI_count_filtered_file,
            sep="\t",
            index=False
        )


def mapping_vdj(args):
    with Mapping_vdj(args, display_title="Mapping") as runner:
        runner.run()


def get_opts_mapping_vdj(parser, sub_program):
    parser.add_argument('--ref_path', help='reference path for igblast')
    parser.add_argument("--type", help='TCR or BCR', required=True)
    parser.add_argument(
        '--species',
        choices=['human', 'mouse'],
        help='Default human. human or mouse.',
        default='human'
    )
    parser.add_argument('--split_fasta', help='split fasta file to map to avoid running out of memory.', action='store_true')
    if sub_program:
        parser.add_argument(
            "--fasta",
            help="Required. Input fasta file.",
            required=True,
        )
        parser = s_common(parser)


import sys

import numpy as np
import pandas as pd

from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.__init__ import HELP_DICT
from celescope.tools.matrix import CountMatrix, Features
from celescope.tools.analysis_wrapper import read_tsne


TAG_COL = 'tag_name'

class Count_cite(Step):

    def __init__(self, args, display_title):
        super().__init__(args, display_title)

        self.df_read_count = pd.read_csv(args.read_count_file, sep='\t', index_col=[0,1])

        self.match_dict = utils.parse_match_dir(args.match_dir)
        self.match_barcode = self.match_dict['match_barcode']
        self.match_matrix_dir = self.match_dict['matrix_dir']
        self.df_rna_tsne = pd.DataFrame()
        if 'tsne_coord' in self.match_dict:
            self.df_rna_tsne = read_tsne(self.match_dict['tsne_coord'])
        else:
            sys.stderr.write('rna tsne file not found!')


        # out
        self.mtx = f'{self.out_prefix}_citeseq.mtx.gz'
        self.matrix_dir = f'{self.out_prefix}_rna_citeseq_matrix'
        self.tsne_file = f'{self.out_prefix}_tsne_coord.tsv'

    @utils.add_log
    def run(self):
        mapped_read = int(self.df_read_count['read_count'].sum())

        # in cell
        df_read_count_in_cell = self.df_read_count[self.df_read_count.index.isin(self.match_barcode, level=0)]
        mapped_read_in_cell = int(df_read_count_in_cell['read_count'].sum())
        self.add_metric(
            name='Mapped Reads in Cells',
            value=mapped_read_in_cell,
            total=mapped_read,
        )

        # UMI
        df_UMI_in_cell = df_read_count_in_cell.groupby([
            'barcode', TAG_COL]).agg({'UMI': 'count'})

        df_temp= df_UMI_in_cell.reset_index().pivot(
            index='barcode', columns=TAG_COL, values='UMI')
        df_cell = pd.DataFrame(index=self.match_barcode)
        df_pivot = pd.merge(
            df_cell,
            df_temp,
            how="left",
            left_index=True,
            right_index=True)

        # fillna
        df_pivot.fillna(0, inplace=True)
        df_pivot = df_pivot.astype(int)
        df_UMI_cell_out = df_pivot.T
        df_UMI_cell_out.to_csv(self.mtx, sep='\t', compression='gzip')

        # merge rna matrix
        tag_names = df_read_count_in_cell.index.get_level_values(1).unique()
        features = Features(tag_names)
        citeseq_matrix = CountMatrix.from_dataframe(df_UMI_in_cell, features, barcodes=self.match_barcode, value='UMI')
        rna_matrix = CountMatrix.from_matrix_dir(matrix_dir=self.match_matrix_dir)
        merged_matrix = rna_matrix.concat_by_barcodes(citeseq_matrix)
        merged_matrix.to_matrix_dir(self.matrix_dir)

        # UMI
        UMIs = df_pivot.apply(sum, axis=1)
        median_umi = round(np.median(UMIs), 2)
        mean_umi = round(np.mean(UMIs), 2)
        self.add_metric(
            name='Median UMI per Cell',
            value=float(median_umi),
        )
        self.add_metric(
            name='Mean UMI per Cell',
            value=float(mean_umi),
        )

        # out tsne
        if not self.df_rna_tsne.empty:
            df_log1p = np.log2(df_pivot + 1)
            df_tsne = self.df_rna_tsne.merge(df_log1p, left_index=True, right_index=True)
            df_tsne.to_csv(self.tsne_file,sep='\t')


def get_opts_count_cite(parser, sub_program):
    if sub_program:
        parser.add_argument("--match_dir", help=HELP_DICT['match_dir'], required=True)
        parser.add_argument("--read_count_file", help="tag read count file")
        s_common(parser)


def count_cite(args):

    with Count_cite(args, display_title='Count') as runner:
        runner.run()




from celescope.tools.capture.filter import Filter, get_opts_filter


def get_opts_filter_fusion(parser, sub_program):

    get_opts_filter(parser, sub_program)

def filter_fusion(args):

    with Filter_fusion(args) as runner:
        runner.run()


class Filter_fusion(Filter):
    """    
    ## Features
    - Correct single-base errors in UMIs due to sequencing, amplification, etc.
    - Filter background UMIs base on a UMI threshold.
    There are three methods to determine the UMI threshold:
        - 'auto' : Using a method similar to cell calling method.
        - 'otsu' : UMI counts are first log 2 transformed and then the threshold is determined by [Otsu's method](https://en.wikipedia.org/wiki/Otsu%27s_method)
        - 'hard' : Using User provided UMI threshold.

    ## Output
    - `{sample}_corrected_read_count.json` Read counts after UMI correction.
    - `{sample}_filtered_read_count.json` Filtered read counts.
    - `{sample}_filtered_UMI.csv` Filtered UMI counts.

    """


import setuptools
from celescope.__init__ import __VERSION__, ASSAY_LIST

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as fp:
    install_requires = fp.read()

entrys = ['celescope=celescope.celescope:main',]
for assay in ASSAY_LIST:
    entrys.append(f'multi_{assay}=celescope.{assay}.multi_{assay}:main')
entry_dict = {
        'console_scripts': entrys,
}


setuptools.setup(
    name="celescope",
    version=__VERSION__,
    author="zhouyiqi",
    author_email="zhouyiqi@singleronbio.com",
    description="Single Cell Analysis Pipelines",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/singleron-RD/CeleScope",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    include_package_data=True,
    entry_points=entry_dict,
    install_requires=install_requires,
)



STEPS = [
    'mkref',
    'sample',
    'barcode',
    'cutadapt',
    'star',
    'prep_map',
    "featureCounts",
    "count"]
__ASSAY__ = 'bulk_rna'

REMOVE_FROM_MULTI = {
    'mkref',
    'prep_map',
}

IMPORT_DICT = {
    'mkref': 'celescope.rna',
    'star': 'celescope.rna'
}



import glob
import gzip
import argparse
import os

import pysam

from celescope.tools import utils
from celescope.__init__ import HELP_DICT
from celescope.celescope import ArgFormatter


SAMPLE = 'test1'

class Extract_read():
    """
    Extract fq1 and fq2 reads according to barcode.
    """

    def __init__(self, args):
        self.args = args
        self.rna_fq_file = glob.glob(f'{args.match_dir}/*barcode/*_2.fq*')[0]
        self.barcodes, _num = utils.read_one_col(args.barcode_file)
        self.read_index = set()

        # out
        self.out_fq1 = f'{args.outdir}/{SAMPLE}_1.fq.gz'
        self.out_fq2 = f'{args.outdir}/{SAMPLE}_2.fq.gz'

        # mkdir
        if not os.path.exists(args.outdir):
            os.system(f'mkdir -p {args.outdir}')


    @utils.add_log
    def write_r2_fastq_files(self):
        read_num = 0
        with pysam.FastxFile(self.rna_fq_file, 'r') as rna_fq:
            with gzip.open(self.out_fq2, 'wt') as fq2:
                for read in rna_fq:
                    read_num += 1
                    attr = read.name.strip("@").split(':')
                    barcode = attr[0]
                    read_index = int(attr[2])
                    if barcode in self.barcodes:
                        self.read_index.add(read_index)
                        fq2.write(str(read) + '\n')

                    if read_num % 1000000 == 0:
                        self.write_r2_fastq_files.logger.info(f'{read_num} done')


    @utils.add_log
    def write_r1_fastq_files(self):
        with pysam.FastxFile(self.args.R1_read, 'r') as r1_read:
            with gzip.open(self.out_fq1, 'wt') as fq1:
                for read_index, read in enumerate(r1_read, start=1):
                    if read_index in self.read_index:
                        fq1.write(str(read) + '\n')

    def run(self):
        self.write_r2_fastq_files()
        self.write_r1_fastq_files()


def main():
    parser = argparse.ArgumentParser(description='Extract fq1 and fq2 reads according to barcode.', formatter_class=ArgFormatter) 
    parser.add_argument("--barcode_file", help="file with barcode to extract", required=True)
    parser.add_argument("--match_dir", help=HELP_DICT['match_dir'], required=True)
    parser.add_argument("--R1_read", help='R1 read path.')
    parser.add_argument("--outdir", help=HELP_DICT['outdir'], default='./')

    args = parser.parse_args()
    runner = Extract_read(args)
    runner.run()

if __name__ == '__main__':
    main()


from collections import defaultdict
import itertools

import pysam
import pandas as pd

from celescope.tools.featureCounts import correct_umi, discard_read
from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.tools import reference
from celescope.tools.matrix import CountMatrix

class Count(Step):
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)
        self.bam = args.bam
        gp = reference.GtfParser(args.gtf)
        gp.get_id_name()
        self.features = gp.get_features()

        self.count_detail_file = f'{self.outdir}/{self.sample}_count_detail.txt'
        self.matrix_dir = f'{self.outdir}/{self.sample}_virus_matrix'

    @staticmethod
    def get_barcodes(filter_umi_file):
        df = pd.read_csv(filter_umi_file)
        return list(df['barcode'])

    def run(self):
        self.bam2table()
        df = pd.read_table(self.count_detail_file, header=0)
        barcodes = Count.get_barcodes(self.args.filter_umi_file)
        df = df.groupby(['Barcode', 'geneID']).agg({"UMI":"count"})

        count_matrix = CountMatrix.from_dataframe(df, self.features, barcodes=barcodes, value="UMI")
        count_matrix.to_matrix_dir(self.matrix_dir)


    @utils.add_log
    def bam2table(self):
        """
        bam to detail table
        must be used on name_sorted bam
        """
        samfile = pysam.AlignmentFile(self.bam, "rb")
        with open(self.count_detail_file, 'wt') as fh1:
            fh1.write('\t'.join(['Barcode', 'geneID', 'UMI', 'count']) + '\n')

            def keyfunc(x):
                return x.query_name.split(':', maxsplit=1)[0]
            for _, g in itertools.groupby(samfile, keyfunc):
                gene_umi_dict = defaultdict(lambda: defaultdict(int))
                for seg in g:
                    (barcode, umi) = seg.query_name.split(':')[:2]
                    if not seg.has_tag('XT'):
                        continue
                    gene_id = seg.get_tag('XT')
                    gene_umi_dict[gene_id][umi] += 1
                for gene_id in gene_umi_dict:
                    correct_umi(gene_umi_dict[gene_id])

                discard_umi, umi_gene_dict = discard_read(gene_umi_dict)

                # output
                for umi in umi_gene_dict:
                    if umi not in discard_umi:
                        for gene_id in umi_gene_dict[umi]:
                            fh1.write('%s\t%s\t%s\t%s\n' % (barcode, gene_id, umi,
                                                            umi_gene_dict[umi][gene_id]))
        samfile.close()

@utils.add_log
def count(args):
    if args.gtf:
        with Count(args, display_title="Cells") as runner:
            runner.run()


def get_opts_count(parser, sub_program):
    parser.add_argument(
        "--gtf",
        help="Optional. Genome gtf file. Use absolute path or relative path to `genomeDir`.",
        )
    if sub_program:
        parser = s_common(parser)
        parser.add_argument('--filter_umi_file', help='filter umi file', required=True)
        parser.add_argument('--bam', help='Required. BAM file from featureCounts.', required=True)




import os
import celescope.tools


STEPS = ['sample', 'barcode', 'mapping', 'assemble', 'summarize', 'annotation']
__ASSAY__ = 'flv_trust4'

TOOLS_DIR = os.path.dirname(celescope.tools.__file__) + '/trust4'
REF_DIR = f'{TOOLS_DIR}/database'

CHAIN = {
	'TCR': ['TRA', 'TRB'], 
	'BCR': ['IGH', 'IGL', 'IGK']
	}
PAIRED_CHAIN = {
	'TCR': ['TRA_TRB'],
	'BCR': ['IGK_IGH', 'IGL_IGH']
}


import json

import pandas as pd

from celescope.tools import utils
from celescope.tools.featureCounts import correct_umi
from celescope.tools.step import Step, s_common
from celescope.tools.capture.threshold import Threshold
from celescope.__init__ import HELP_DICT
from celescope.tools.capture.__init__ import SUM_UMI_COLNAME



def get_opts_filter(parser, sub_program):
    parser.add_argument('--not_correct_UMI', help='Do not perform UMI correction.', action='store_true')

    parser.add_argument(
        "--read_threshold_method",
        help='method to find read threshold. UMIs with `support reads` < `read threshold` are filtered.',
        choices=['otsu', 'auto', 'hard', 'none'],
        default='otsu'
    )
    parser.add_argument(
        "--read_hard_threshold",
        help='int, use together with `--read_threshold_method hard`',
    )

    parser.add_argument(
        "--umi_threshold_method",
        help='method to find UMI threshold. Cell barcode with `UMI` < `UMI threshold` are considered negative.',
        choices=['otsu', 'auto', 'hard', 'none'],
        default='otsu'
    )
    parser.add_argument(
        "--umi_hard_threshold",
        help='int, use together with `--umi_threshold_method hard`',
    )
    parser.add_argument(
        "--auto_coef",
        help='int, threshold = top 1 percent positive cell count / auto_coef',
        default=3,
    )
    parser.add_argument(
        "--otsu_log_base",
        help='raw counts are first log transformed before thresholding. This argument is the log base. Commonly used values are 2 and 10.',
        default=10,
        type=int,
    )


    if sub_program:
        parser.add_argument('--match_dir', help=HELP_DICT['match_dir'], required=True)
        parser.add_argument('--raw_read_count_file', required=True)
        s_common(parser)


class Filter(Step):
    """    
    ## Features
    - Correct single-base errors in UMIs due to sequencing, amplification, etc.
    - Filter background UMIs base on a UMI threshold.
    There are three methods to determine the UMI threshold:
        - 'auto' : Using a method similar to cell calling method.
        - 'otsu' : UMI counts are first log transformed and then the threshold is determined by [Otsu's method](https://en.wikipedia.org/wiki/Otsu%27s_method)
        - 'hard' : Using User provided UMI threshold.

    ## Output
    - `{sample}_corrected_read_count.json` Read counts after UMI correction.
    - `{sample}_filtered_read_count.json` Filtered read counts.
    - `{sample}_filtered_UMI.csv` Filtered UMI counts.

    """
    def __init__(self, args, display_title='Filtering'):
        super().__init__(args, display_title)

        # data
        with open(args.raw_read_count_file) as f:
            self.count_dict = json.load(f)

        self.raw_umi = 0
        self.total_corrected_umi = 0
        self.del_umi = 0
        self.read_threshold_dict = {}
        self.umi_threshold_dict = {}  # if not set explicitly, use 1 as default

        self.barcode_ref_umi_dict = utils.genDict(dim=2)
        self.ref_barcode_umi_dict = utils.genDict(dim=2)

        match_dir_dict = utils.parse_match_dir(args.match_dir)
        self.match_barcode = match_dir_dict['match_barcode']
        self.df_filter_umi = pd.DataFrame(index=list(self.match_barcode)).rename_axis('barcode')

        # out
        self.corrected_read_count_file = f'{self.out_prefix}_corrected_read_count.json'
        self.filter_read_count_file = f'{self.out_prefix}_filtered_read_count.json'
        self.filter_umi_file = f'{self.out_prefix}_filtered_UMI.csv'

    @utils.add_log
    def correct_umi(self):
        for barcode in self.count_dict:
            for ref in self.count_dict[barcode]:
                self.raw_umi += len(self.count_dict[barcode][ref])
                n_corrected_umi, _n_corrected_read, _ = correct_umi(self.count_dict[barcode][ref])
                if self.debug:
                    print(f'{barcode} {ref} {n_corrected_umi}')
                self.total_corrected_umi += n_corrected_umi

        self.add_metric(
            name='Number of Raw UMI',
            value=self.raw_umi,
            help_info='number of total raw UMI',
        )

        self.add_metric(
            name='Number of Corrected UMI',
            value=self.total_corrected_umi,
            total=self.raw_umi,
            help_info='correct sequencing errors in the UMI sequences ',
        )

    @utils.add_log
    def write_correct_umi_json(self):
        with open(self.corrected_read_count_file, 'w') as fp:
            json.dump(self.count_dict, fp, indent=4)

    @utils.add_log
    def get_read_threshold(self):
        self.add_metric(
            'Read Threshold Method', 
            self.args.read_threshold_method,
            help_info=f"""There are three methods to determine the threshold:<br>
1. 'auto' : Using a method similar to cell calling method. threshold = top 1% positive cell count / auto_coef. auto_coef = {self.args.auto_coef} <br>
2. 'otsu' : UMI counts are first log transformed and then the threshold is determined by Otsu's method. otsu_log_base = {self.args.otsu_log_base} <br>
3. 'hard' : Using User provided UMI threshold."""
        )

        if self.args.read_threshold_method == 'auto':
            self.add_metric(
                name='Read auto coeffient',
                value=self.args.auto_coef,
                help_info='threshold = top 1% positive cell count / auto_coef',
            )

        read_dict = utils.genDict(dim=1, valType=list)

        for barcode in self.count_dict:
            for ref in self.count_dict[barcode]:
                read_dict[ref] += list(self.count_dict[barcode][ref].values())

        if self.debug:
            print(read_dict)

        for ref in read_dict:              
            otsu_plot_path = f'{self.out_prefix}_{ref}_read_otsu.png'
            runner = Threshold(
                array=read_dict[ref], 
                threshold_method=self.args.read_threshold_method, 
                otsu_plot_path=otsu_plot_path,
                hard_threshold=self.args.read_hard_threshold,
                coef=self.args.auto_coef,
                log_base=self.args.otsu_log_base,
            )
            read_threshold = runner.run()
            
            self.read_threshold_dict[ref] = read_threshold
            self.add_metric(
                f'{ref} Read Threshold', 
                read_threshold,
            )


    @utils.add_log
    def filter_read(self):
        for barcode in self.count_dict:
            for ref in self.count_dict[barcode]:
                for umi in self.count_dict[barcode][ref]:
                    read_count = self.count_dict[barcode][ref][umi]
                    if read_count < self.read_threshold_dict[ref]:
                        self.del_umi += 1
                        self.count_dict[barcode][ref][umi] = 0

        """
        self.add_metric(
            name='Number of Total Filtered UMI',
            value=self.del_umi,
            total=self.raw_umi,
        )
        """

    def write_filter_read_json(self):
        with open(self.filter_read_count_file, 'w') as fp:
            json.dump(self.count_dict, fp, indent=4)


    @utils.add_log
    def set_barcode_ref_umi_dict(self):

        for barcode in self.count_dict:
            for ref in self.count_dict[barcode]:
                for umi in self.count_dict[barcode][ref]:
                    if self.count_dict[barcode][ref][umi] > 0:
                        self.barcode_ref_umi_dict[barcode][ref] += 1
        if self.debug:
            print(dict(self.barcode_ref_umi_dict))

    @utils.add_log
    def set_ref_barcode_umi_dict(self):

        for barcode in self.barcode_ref_umi_dict:
            for ref in self.barcode_ref_umi_dict[barcode]:
                self.ref_barcode_umi_dict[ref][barcode] = self.barcode_ref_umi_dict[barcode][ref]

    def get_umi_threshold(self):
        self.add_metric(
            'UMI Threshold Method', 
            self.args.umi_threshold_method,
            help_info="""Use the same threshold method as read count threshold."""
        )

        if self.args.umi_threshold_method == 'auto':
            self.add_metric(
            name='UMI auto coeffient',
            value=self.args.auto_coef,
            help_info='threshold = top 1% positive cell count / auto_coef',
        )

        for ref in self.ref_barcode_umi_dict:
            umi_array = list(self.ref_barcode_umi_dict[ref].values())
            otsu_plot_path = f'{self.out_prefix}_{ref}_UMI_otsu.png'
            runner = Threshold(
                umi_array, 
                threshold_method=self.args.umi_threshold_method, 
                otsu_plot_path=otsu_plot_path,
                hard_threshold=self.args.umi_hard_threshold,
                coef=self.args.auto_coef,
                log_base=self.args.otsu_log_base,
            )
            umi_threshold = runner.run()
            
            umi_threshold = max(1, umi_threshold)
            self.umi_threshold_dict[ref] = umi_threshold
            self.add_metric(f'{ref} UMI Threshold', umi_threshold)


    @utils.add_log
    def filter_umi(self):
        for ref in self.ref_barcode_umi_dict:
            for barcode in self.ref_barcode_umi_dict[ref]:
                if self.ref_barcode_umi_dict[ref][barcode] < self.umi_threshold_dict[ref]:
                    self.ref_barcode_umi_dict[ref][barcode] = 0

    @utils.add_log
    def add_umi_write_csv(self):
        for ref in self.umi_threshold_dict:
            self.df_filter_umi[ref] = pd.Series(self.ref_barcode_umi_dict[ref])
            self.df_filter_umi[ref].fillna(0, inplace=True)

        refs = list(self.umi_threshold_dict.keys()) 
        self.df_filter_umi[SUM_UMI_COLNAME] = self.df_filter_umi[refs].sum(axis=1)
        self.df_filter_umi.to_csv(self.filter_umi_file)


    def add_some_metrics(self):
        df = self.df_filter_umi

        cell_total = len(df)
        df_positive = df[df[SUM_UMI_COLNAME] > 0]
        n_cell_positive = len(df_positive)
        self.add_metric(
            name='Number of Positive Cells after Filtering',
            value=n_cell_positive,
            total=cell_total,
            help_info='Cells with sum_UMI > 0 after filtering',
        )

    def run(self):
        if not self.args.not_correct_UMI:
            self.correct_umi()
            self.write_correct_umi_json()

        self.get_read_threshold()
        self.filter_read()
        self.write_filter_read_json()

        self.set_barcode_ref_umi_dict()
        self.set_ref_barcode_umi_dict()
        self.get_umi_threshold()
        self.filter_umi()

        self.add_umi_write_csv()
        self.add_some_metrics()





from celescope.tools import utils
from celescope.__init__ import __VERSION__
from celescope.tools.__init__ import PATTERN_DICT
from celescope.tools.barcode import Chemistry
from celescope.tools.step import Step, s_common


def add_kit_version(chemistry):
    kit_dict = {
        '1': 'no longer in use',
        '2': 'kit V1',
        '3': 'kit V2',
    }
    if chemistry.startswith('scopeV'):
        s = chemistry.replace('scopeV', '')
        chem_version = s[0]
        kit = kit_dict[chem_version]
        chemistry = f'{chemistry} ({kit})'
    
    return chemistry

class Sample(Step):
    def __init__(self, args):
        Step.__init__(self, args)
        self.version = __VERSION__
        self.chemistry = args.chemistry
        self.wells = args.wells

    @utils.add_log
    def run(self):
        if self.chemistry == 'auto':
            fq1 = self.args.fq1
            ch = Chemistry(fq1, self.assay)
            chemistry = ch.check_chemistry()
            chemistry = ",".join(set(chemistry))
        else:
            chemistry = self.chemistry
        
        if chemistry == 'bulk_rna':
            chemistry = f'accuracode{self.wells}'

        self.add_metric(
            name='Sample ID',
            value=self.sample,
        )
        self.add_metric(
            name='Assay',
            value=self.assay,
        )
        self.add_metric(
            name='Chemistry',
            value=chemistry,
            display=add_kit_version(chemistry),
            help_info='For more information, see <a href="https://github.com/singleron-RD/CeleScope/blob/cbf5df39b74628fbcc1d9265728dcfe86b6989f2/doc/chemistry.md">here</a>',
        )
        self.add_metric(
            name='Software Version',
            value=self.version,
        )


@utils.add_log
def sample(args):

    with Sample(args) as runner:
        runner.run()


def get_opts_sample(parser, sub_program):
    if sub_program:
        parser = s_common(parser)
        parser.add_argument('--fq1', help='read1 fq file')
    parser.add_argument('--chemistry', choices=list(PATTERN_DICT.keys()), help='chemistry version', default='auto')
    parser.add_argument('--wells', type=int,
        help='The AccuraCode wells used (384 or 96), default 384.', default=384)
    return parser



from celescope.sweetseq.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_sweetseq(Multi):
    """
    ## Usage
    Before running `multi_sweetseq`, you need to run scRNA-Seq data with CeleScope first.
    ```
    multi_sweetseq \\
        --mapfile ./sweetseq.mapfile\\
        --barcode_fasta celescope/data/sweetseq/sweet_tag_barcode.fasta\\
        --linker_fasta celescope/data/sweetseq/sweet_tag_linker.fasta \\
        --fq_pattern L23C15\\
        --mod shell
    ```
    """

    def mapping_tag(self, sample):
        step = 'mapping_tag'
        fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '

        )
        self.process_cmd(cmd, step, sample, m=2, x=1)

    def count_tag(self, sample):
        step = 'count_tag'
        read_count_file = f'{self.outdir_dic[sample]["mapping_tag"]}/{sample}_read_count.tsv'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--match_dir {self.col4_dict[sample]} '
            f'--read_count_file {read_count_file} '

        )
        self.process_cmd(cmd, step, sample, m=2, x=1)


    def analysis_tag(self, sample):
        step = 'analysis_tag'
        cmd_line = self.get_cmd_line(step, sample)
        tsne_tag_file = f'{self.outdir_dic[sample]["count_tag"]}/{sample}_tsne_tag.tsv'
        cmd = (
            f'{cmd_line} '
            f'--match_dir {self.col4_dict[sample]} '
            f'--tsne_tag_file {tsne_tag_file} '
        )
        self.process_cmd(cmd, step, sample, m=2, x=1)


def main():
    multi = Multi_sweetseq(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()



from celescope.tools import analysis_wrapper
from celescope.tools.plotly_plot import Tsne_plot
from celescope.tools import utils
from celescope.tools.step import Step


class Analysis(Step):
    """
    ## Features

    - Calculate the marker gene of each cluster.

    - Cell type annotation(optional). You can provide markers of known cell types and annotate cell types for each cluster.

    ## Output
    - `markers.tsv` Marker genes of each cluster.

    - `tsne_coord.tsv` t-SNE coordinates and clustering information.

    - `{sample}/06.analsis/{sample}_auto_assign/` This result will only be obtained when `--type_marker_tsv` 
    parameter is provided. The result contains 3 files:
        - `{sample}_auto_cluster_type.tsv` The cell type of each cluster; if cell_type is "NA", 
    it means that the given marker is not enough to identify the cluster.
        - `{sample}_png/{cluster}_pctdiff.png` Percentage of marker gene expression in this cluster - percentage in all other clusters.
        - `{sample}_png/{cluster}_logfc.png` log2 (average expression of marker gene in this cluster / average expression in all other clusters + 1)
    """

    def __init__(self, args, display_title=None):

        super().__init__(args, display_title)

        self.display_title = display_title
        self._table_id = 'marker_genes'

    def run(self):

        with analysis_wrapper.Scanpy_wrapper(self.args, display_title=self.display_title) as scanpy_wrapper:
            scanpy_wrapper.run()
            df_tsne, df_marker = scanpy_wrapper.get_df()
            self.set_metric_list(metric_list=scanpy_wrapper.get_metric_list())

        with analysis_wrapper.Report_runner(self.args, display_title=self.display_title) as report_runner:
            report_runner.add_marker_help()


        tsne_cluster = Tsne_plot(df_tsne, 'cluster').get_plotly_div()
        self.add_data(tsne_cluster=tsne_cluster)

        tsne_gene = Tsne_plot(df_tsne, 'Gene_Counts', discrete=False).get_plotly_div()
        self.add_data(tsne_gene=tsne_gene)

        table_dict = self.get_table_dict(
            title='Marker Genes by Cluster',
            table_id=self._table_id,
            df_table=df_marker,
        )
        self.add_data(table_dict=table_dict)


@utils.add_log
def analysis(args):
    with Analysis(args, display_title='Analysis') as runner:
        runner.run()


def get_opts_analysis(parser, sub_program):
    analysis_wrapper.get_opts_analysis(parser, sub_program)
